
#include "general.h"
#include <iostream.h>
#include "lifetime_polka.H"

#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/ArrowB.h>
#include <Xm/ScrollBar.h>


/***************************************************************/
/**********	    global variables		      **********/
/***************************************************************/

extern   int           _delay;
extern   Display      *_display;
extern   XtAppContext  _context;

void              PolkaInit();
void              BatchMode(int);
void              CheckXEvents();
void              micro_sleep(int);

/******  In polkawidget.C   ******/
void              resize(Widget, XtPointer, XtPointer);
void              tlresize(Widget, XtPointer, XtPointer);
void              clrmapresize(Widget, XtPointer, XtPointer);

/******  In polkagc.C  ********/
void              SetBackground(COLORINDEX);
COLORINDEX        load_color(COLOR);
Pixel             GetXColor(COLORINDEX);
COLORINDEX        set_color(COLORINDEX);
LINE_STYLE        line_style(LINE_STYLE);
FILL_STYLE        fill_style(FILL_STYLE);
int               load_font(char *);
void              text_info(Font, char *, int *, int *, int *, int *);
int               set_font(int);
GC                inq_gc();
int               inq_color();
int               inq_font();
void              initGC();

/******  In polkaobject.C  ********/
FILL_STYLE	  get_fillstyle(double);

extern Widget		_appshell;

extern void debugCB(Widget, XtPointer, XtPointer);
extern void refreshCB(Widget, XtPointer, XtPointer);
extern void closeCB(Widget, XtPointer, XtPointer);

static void panlCB(Widget, XtPointer, XtPointer);
static void panrCB(Widget, XtPointer, XtPointer);
static void pandCB(Widget, XtPointer, XtPointer);
static void panuCB(Widget, XtPointer, XtPointer);
static void zoominCB(Widget, XtPointer, XtPointer);
static void zoomoutCB(Widget, XtPointer, XtPointer);
static void vzoominCB(Widget, XtPointer, XtPointer);
static void vzoomoutCB(Widget, XtPointer, XtPointer);
static double hzoom_factor = 0.5, vzoom_factor = 0.5;
static double hmove_factor = 0.5, vmove_factor = 0.5;

extern APP_CALLBACK AppCloseCB;

extern void PolkaInit();

extern int debug_flag;

 
/***************************************************************/
/**********	  constant definitions		      **********/
/***************************************************************/

#define   WIDTH             500
#define   HEIGHT            500
#define   MAXSTRINGLENGTH   256
#define   ZERO              0
#define   MYPI              ((double)3.14159)

#define WHITE             0
#define YELLOW            1
#define GREEN             2
#define BLUE              3
#define ORANGE            4
#define RED               5
#define MAROON            6
#define BLACK             7

#define   POLKA_BASE_FILL         0
#define   POLKA_NUM_FILL         41
#define   POLKA_FILL_OUTLINE     POLKA_BASE_FILL
#define   POLKA_FILL_SOLID       POLKA_BASE_FILL + POLKA_NUM_FILL - 1


TimedView::TimedView(double h_zoom, double v_zoom,
		     double h_move, double v_move,
		     ResizeMode rm, double start_time,
		     double screen_time, int ticks) : View()
{
    timelinewin = clrmapwin = 0;
    start_ts = 0.0;
    screen_length = 2.0;
    screen_ticks = 20;
    hzoom_factor = h_zoom_ratio = h_zoom;
    vzoom_factor = v_zoom_ratio = v_zoom;
    hmove_factor = h_move_ratio = h_move;
    vmove_factor = v_move_ratio = v_move;
    resize_mode = rm;
    start_ts = start_time; 
    screen_length = screen_time;
    screen_ticks = ticks;
}


/*---------------------------------------------------------------------------*/
/*                              POLKA Widgets                                */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*                             toplevel                                      */
/*                                |                                          */
/*                               form                                        */
/*       /---------------/--------+---\-------\-----\-----\------\-----\     */
/*     title           rcleft         sbar   pause mode debug refresh quit   */
/*               /----/---\---\---\--\                                       */
/*            left right up down in out        			             */
/*                                                                           */
/*                                                                           */
/*	modeMgrShell			debugMgrShell		      */
/*	      |				      |			      */
/*	modeMenuMgr			debugMenuMgr		      */
/*	      |				      |			      */
/*	 modeShell			 debugShell		      */
/*	      |				      |			      */
/*	  Refresh			    Debug		      */
/*    /---------|-------\		    /-+-\		      */
/* by frame  by scene  none		  *Off  On		      */
/*---------------------------------------------------------------------------*/

/**************************************************************/
/* setupWindow -- Create, init (hardcoded), place, &          */
/*		  realize all widgets for animation windows.  */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
TimedView::setupWindow(char *title)
{
    Widget      form, botform, botrightrowcol, debugbut,
    		refreshbut, closebut,
                botleftrowcol, larrowbut, rarrowbut, uarrowbut, darrowbut, 
                zoominbut, zoomoutbut;
    XmString    xmstr;
    Arg	        args[20];
    int         n;
    
    if (debug)
      printf("DEBUG: View::setupWindow\n");

    n = 0;
    XtSetArg(args[n], XmNtitle, title); n++;
    topshell = XtCreatePopupShell("AnimWindow", topLevelShellWidgetClass, 
				  _appshell, args, n);


    form = XtCreateManagedWidget("topform", xmFormWidgetClass, topshell,   
				 NULL, 0);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    botform =
      XtCreateManagedWidget("botform", xmFormWidgetClass, form, args, n);

    n = 0;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
    botleftrowcol =
      XtCreateManagedWidget("botleftrowcol", xmRowColumnWidgetClass,
			    botform, args, n);

    n = 0;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
    botrightrowcol =
      XtCreateManagedWidget("botrightrowcol", xmRowColumnWidgetClass,
			    botform, args, n);

    n = 0;
    xmstr = XmStringCreate((char*)"Debug",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    XtSetArg(args[n], XmNindicatorOn, False); n++;
    debugbut =
      XtCreateManagedWidget("debugbut", xmToggleButtonWidgetClass,
			    botrightrowcol, args, n);
    XtAddCallback(debugbut, XmNvalueChangedCallback, debugCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"Refresh",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    refreshbut =
      XtCreateManagedWidget("modebut", xmPushButtonWidgetClass, 
			    botrightrowcol, args, n);
    XtAddCallback(refreshbut, XmNactivateCallback, refreshCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"Close",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    closebut =
      XtCreateManagedWidget("closebut", xmPushButtonWidgetClass,
                            botrightrowcol, args, n);
    XtAddCallback(closebut, XmNactivateCallback, closeCB, (XtPointer)topshell);
    if (AppCloseCB) {
	XtAddCallback(closebut, XmNactivateCallback, *AppCloseCB,
		      (XtPointer)this);
    }

    n = 0;
    XtSetArg(args[n], XmNarrowDirection, XmARROW_LEFT); n++;
    larrowbut =
      XtCreateManagedWidget("larrowbut", xmArrowButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(larrowbut, XmNactivateCallback, panlCB, (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XmNarrowDirection, XmARROW_RIGHT); n++;
    rarrowbut =
      XtCreateManagedWidget("rarrowbut", xmArrowButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(rarrowbut, XmNactivateCallback, panrCB, (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XmNarrowDirection, XmARROW_DOWN); n++;
    darrowbut =
      XtCreateManagedWidget("darrowbut", xmArrowButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(darrowbut, XmNactivateCallback, pandCB, (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XmNarrowDirection, XmARROW_UP); n++;
    uarrowbut =
      XtCreateManagedWidget("uarrowbut", xmArrowButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(uarrowbut, XmNactivateCallback, panuCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"h-In",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    zoominbut =
      XtCreateManagedWidget("zoominbut", xmPushButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(zoominbut, XmNactivateCallback, zoominCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"h-Out",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    zoomoutbut =
      XtCreateManagedWidget("zoomoutbut", xmPushButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(zoomoutbut, XmNactivateCallback, zoomoutCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"v-In",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    zoominbut =
      XtCreateManagedWidget("vzoominbut", xmPushButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(zoominbut, XmNactivateCallback, vzoominCB, (XtPointer)this);

    n = 0;
    xmstr = XmStringCreate((char*)"v-Out",XmSTRING_DEFAULT_CHARSET); 
    XtSetArg(args[n], XmNlabelString, xmstr); n++;
    zoomoutbut =
      XtCreateManagedWidget("vzoomoutbut", xmPushButtonWidgetClass, 
			    botleftrowcol, args, n);
    XtAddCallback(zoomoutbut, XmNactivateCallback, vzoomoutCB, (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XtNbackground,
	     WhitePixel(_display, DefaultScreen(_display))); n++;
    XtSetArg(args[n], XtNborderWidth, 1); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNbottomWidget, botform); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNwidth, 110); n++;
    clrmapwid =
      XtCreateManagedWidget("colormapping", xmDrawingAreaWidgetClass, form,
			    args, n);
    XtAddCallback(clrmapwid, XmNexposeCallback, clrmapresize,
		  (XtPointer)this);
    XtAddCallback(clrmapwid, XmNresizeCallback, clrmapresize,
		  (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XtNborderWidth, 1); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNbottomWidget, botform); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, clrmapwid); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNheight, 40); n++;
    timelinewid =
      XtCreateManagedWidget("timeline", xmDrawingAreaWidgetClass, form,
			    args, n);
    XtAddCallback(timelinewid, XmNexposeCallback, tlresize, (XtPointer)this);
    XtAddCallback(timelinewid, XmNresizeCallback, tlresize, (XtPointer)this);

    n = 0;
    XtSetArg(args[n], XtNborderWidth, 1); n++;
    XtSetArg(args[n], XmNresizePolicy, XmRESIZE_ANY); n++;
    XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNbottomWidget, timelinewid); n++;
    XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
    XtSetArg(args[n], XmNleftWidget, clrmapwid); n++;
    XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
    XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
    if ((pixX > 0) && (pixY > 0)) {   // set by the user via View::Create
	XtSetArg(args[n], XmNwidth, pixX); n++;
	XtSetArg(args[n], XmNheight, pixY); n++;
    }
    easel =
      XtCreateManagedWidget("animwindow", xmDrawingAreaWidgetClass, form,
			    args, n);
    XtAddCallback(easel, XmNexposeCallback, resize, (XtPointer)this);
    XtAddCallback(easel, XmNresizeCallback, resize, (XtPointer)this);
    XtAddEventHandler(easel, ButtonPressMask, FALSE, SelectPoint, 
		      (XtPointer)this);

    XtPopup(topshell, XtGrabNone);
}

int
TimedView::Create(const char *title, int wid, int hei,
		  const char *, const char *)
{
    Window w;
    AOPtr walker,curr;

    PolkaInit();   // just to be safe

    pixX = wid;	 // If these are 0 here, they will be overriden
    pixY = hei;	 // in setupWindow

    if (!created) {
	setupWindow((char*)title);
		// sets up X window and Xt widgets
	timevalFont = XLoadFont(_display, "8x13");
	tslabelFont = XLoadFont(_display, "9x15");
	legendFont = XLoadFont(_display, "6x13");
	timevalFontStruct = XLoadQueryFont(_display, "8x13");
	tslabelFontStruct = XLoadQueryFont(_display, "9x15");
	legendFontStruct = XLoadQueryFont(_display, "6x13");
	timelinewin = XtWindow(timelinewid);
	clrmapwin = XtWindow(clrmapwid);
	DrawTimeLineTicks();
	DrawTimeValues();
    }

    /* Do default values */
    resizeMode = resize_mode;
    bgcolor = WHITE;
    numobjects = 0;
    motionblur = 0;
    if (created) {        // need to remove old stuff
	walker = AOTail;
	while (walker) {
	    curr = walker;
	    walker = walker->prev;
	    delete(curr);
	}
	walker = PendingTail;
	while (walker) {
	    curr = walker;
	    walker = walker->prev;
	    delete(curr);
	}
    }
    AOHead = AOTail = NULL;
    PendingHead = PendingTail = NULL;

    if (!created) {
	cout << "Creating\n";
	w = XtWindow(easel);
	prepareAnim(w);  /* sets destwin & window member variables */
	time = 0;   // need to reset
    }

    aspectx = (double)pixX / dx;
    aspecty = (double)pixY / dy;

    fixPad();   // calculates boxpad 

    created = 1;
    return(1);
}

void
TimedView::DrawTimeLineTicks()
{
    int          i;
    Window       root;
    int          x, y;
    unsigned int lx1, ly1, borderwid, depth;
    int xx, height;
    FILL_STYLE   oldfill, newfill;

    if (timelinewin == 0) return;
    XGetGeometry(_display, timelinewin, &root, &x, &y, &lx1, &ly1,
		 &borderwid, &depth);
    newfill = get_fillstyle(1.0);
    oldfill = fill_style(newfill);
    set_color(BLACK);
    line_style(SOLID);
    for (i = 0; i <= screen_ticks; i++) {
	xx = (i * lx1) / screen_ticks;
	if (i % 10 == 0) {
	    height = 10;
	} else if (i % 5 == 0) {
	    height = 8;
	} else {
	    height = 5;
	}
	XDrawLine(_display, timelinewin, inq_gc(), xx, 0, xx, height);
    }
    fill_style(oldfill);
}

void
TimedView::DrawTimeValues()
{
    Window       root;
    int          x, y, width;
    unsigned int lx1, ly1, borderwid, depth;
    char         buf[40];
    double       factor;
    FILL_STYLE   oldfill, newfill;

    if (timelinewin == 0) return;
    XGetGeometry(_display, timelinewin, &root, &x, &y, &lx1, &ly1,
		 &borderwid, &depth);
    newfill = get_fillstyle(1.0);
    oldfill = fill_style(newfill);
    set_color(BLACK);
    XSetFont(_display, inq_gc(), timevalFont);
    if (screen_length >= 1.0) {
	factor = 1.0;
    } else if (screen_length >= 0.001) {
	factor = 1000.0;
    } else {
	factor = 1000000.0;
    }

    XClearArea(_display, timelinewin, 0, 11, lx1, 10, False);
    sprintf(buf, "%.1f", start_ts*factor);
    XDrawString(_display, timelinewin, inq_gc(), 2, 20, buf, strlen(buf));
    sprintf(buf, "%.1f", (start_ts+screen_length*0.5)*factor);
    width = XTextWidth(timevalFontStruct, buf, strlen(buf));
    XDrawString(_display, timelinewin, inq_gc(), (lx1-width)/2, 20,
		buf, strlen(buf));
    sprintf(buf, "%.1f", (start_ts+screen_length)*factor);
    width = XTextWidth(timevalFontStruct, buf, strlen(buf));
    XDrawString(_display, timelinewin, inq_gc(), lx1-width-2, 20,
		buf, strlen(buf));
    fill_style(oldfill);
}

void
TimedView::DrawTSLabel()
{
    Window       root;
    int          x, y, width;
    unsigned int lx1, ly1, borderwid, depth;
    char         buf[16], tsunit[16];

    if (timelinewin == 0) return;
    XGetGeometry(_display, timelinewin, &root, &x, &y, &lx1, &ly1,
		 &borderwid, &depth);
    set_color(BLACK);
    XSetFont(_display, inq_gc(), tslabelFont);
    if (screen_length >= 1.0) {
	strcpy(tsunit, "(seconds)");
    } else if (screen_length >= 0.001) {
	strcpy(tsunit, "(milliseconds)");
    } else {
	strcpy(tsunit, "(microseconds)");
    }

    XClearArea(_display, timelinewin, 0, 21, lx1, 15, False);
    strcpy(buf, "execution time");
    width = XTextWidth(tslabelFontStruct, buf, strlen(buf));
    XDrawString(_display, timelinewin, inq_gc(), lx1/2 - width -2, 35,
		buf, strlen(buf));
    XDrawString(_display, timelinewin, inq_gc(), lx1/2+2, 35,
		tsunit, strlen(tsunit));
}

void
TimedView::DrawLegend()
{
    int  left_margin = 5, top_margin = 10, height = 45;
    int  rect_x = 0, rect_y = 0;
    unsigned int  rect_h = 20, rect_w = 90;
    int  str_x = 0, str_y = 30;
    const char *str1 = "thread running";
    const char *str2 = "in ready queue";
    const char *str3 = "waiting mutex";
    const char *str4 = "waiting condition";
    const char *str5 = "waiting to join";
    const char *str6 = "to be joined";
    const char *str7 = "perturbation";
    FILL_STYLE  oldfill;
    
    if (clrmapwin == 0) return;
    XSetFont(_display, inq_gc(), legendFont);
#ifdef BLACK_WHITE
    set_color(BLACK);
#else
    set_color(GREEN);
#endif
    oldfill = fill_style(get_fillstyle(1.0));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 0 * height + rect_y,
		   rect_w, rect_h);
    fill_style(get_fillstyle(0.4));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 1 * height + rect_y,
		   rect_w, rect_h);
    set_color(BLACK);
    fill_style(get_fillstyle(0.2));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 2 * height + rect_y,
		   rect_w, rect_h);
    fill_style(get_fillstyle(0.5));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 3 * height + rect_y,
		   rect_w, rect_h);
    fill_style(get_fillstyle(0.8));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 4 * height + rect_y,
		   rect_w, rect_h);
    set_color(BLACK);
    fill_style(get_fillstyle(0.1));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 5 * height + rect_y,
		   rect_w, rect_h);
#ifndef NO_PERTURBATION
#ifdef BLACK_WHITE
    set_color(BLACK);
#else
    set_color(RED);
#endif
    fill_style(get_fillstyle(0.65));
    XFillRectangle(_display, clrmapwin, inq_gc(),
		   left_margin + rect_x, top_margin + 6 * height + rect_y,
		   rect_w, rect_h);
#endif

    set_color(BLACK);
    fill_style(get_fillstyle(1.0));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 0 * height + str_y,
		str1, strlen(str1));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 1 * height + str_y,
		str2, strlen(str2));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 2 * height + str_y,
		str3, strlen(str3));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 3 * height + str_y,
		str4, strlen(str4));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 4 * height + str_y,
		str5, strlen(str5));
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 5 * height + str_y,
		str6, strlen(str6));
#ifndef NO_PERTURBATION
    XDrawString(_display, clrmapwin, inq_gc(),
		left_margin + str_x, top_margin + 6 * height + str_y,
		str7, strlen(str7));
#endif
    fill_style(oldfill);
}

void
TimedView::ClearTimeValueArea()
{
    Window       root;
    int          x, y;
    unsigned int lx1, ly1, borderwid, depth;

    if (timelinewin == 0) return;
    XGetGeometry(_display, timelinewin, &root, &x, &y, &lx1, &ly1,
		 &borderwid, &depth);

    XClearArea(_display, timelinewin, 0, 0, lx1, ly1, True);
}


ThreadLifeTime_polka::ThreadLifeTime_polka(int thrds_in_win,
					   double scrnwidth,
					   double h_zoom, double v_zoom,
					   double h_move, double v_move,
					   int resize_const,
					   int num_of_colors,
					   char **names_of_colors)
	:ThreadLifeTime()
{
    max_num_threads = MAX_NUM_THREADS_INIT;
    h_zoom_ratio = h_zoom;
    v_zoom_ratio = v_zoom;
    h_move_ratio = h_move;
    v_move_ratio = v_move;
    if (resize_const) {
	resize_mode = ConstantAspect;
    } else {
	resize_mode = CoordStretch;
    }
    screen_width = scrnwidth;
    origx = 0.0; origy = 0.0;
    num_threads_in_unit_window = thrds_in_win;
    yincr = (1.0-origy) / thrds_in_win;
    thickness = yincr * 0.6;
    max_seconds = MAX_SECONDS_INIT;
    polkaview = NULL;
    screen_width_units = scrnwidth * timestamp_resolution;
    screen_width_units_inv = 1.0 / screen_width_units;
    xpos = new double[max_num_threads];
    ypos = new double[max_num_threads];
    position_index = new int[max_num_threads];
    max_num_positions = MAX_NUM_THREADS_INIT;
    position_array = new int[max_num_positions];
    next_position = 0;
    global_time = 0;
    polkaview = NULL;
    color_depot = new Palette(num_of_colors, names_of_colors);
}


ThreadLifeTime_polka::~ThreadLifeTime_polka()
{
    delete(xpos);
    delete(ypos);
    delete(color_depot);
    delete(polkaview);
    delete(position_index);
    delete(position_array);
}


void
ThreadLifeTime_polka::lifetimeInitialize()
{
    if (polkaview == NULL) {
        polkaview = new TimedView(h_zoom_ratio, v_zoom_ratio,
				  h_move_ratio, v_move_ratio,
				  resize_mode, 0, screen_width, 20);
    }
    polkaview->Create("Thread Lifetime View");
    screen_width_units = screen_width * timestamp_resolution;
    screen_width_units_inv = 1.0 / screen_width_units;

    //DBG1("timestamp_resolution is %u\n", timestamp_resolution);
    //text = new Text(polkaview, True, 0.5, 0.95, "black", "r24",
    //		    "Lifetime of Threads", 1);
    //text->Originate(global_time);
}


void
ThreadLifeTime_polka::lifetimeNewThread(int thread_idx,
					unsigned long timestamp,
					int child_thread_idx, char*)
{
    int    i, j, oldmax;
    Line   *line;
    double *new_xpos, *new_ypos;
    int    *new_position_index, *new_position_array;

    /* DBG4("New Thread: thread %d, time %u, child %d, name %s\n",
	 thread_idx, timestamp, child_thread_idx, name); */
    if (thread_idx < 0) {
	xpos[child_thread_idx] =
	  (timestamp-init_timestamp) * screen_width_units_inv; /* origx; */
	ypos[child_thread_idx] = origy + yincr*child_thread_idx;
	lifetimeHScroll(0, xpos[child_thread_idx]);
	position_index[child_thread_idx] = 0;
	position_array[0] = child_thread_idx;
	next_position++;
	return;
    }

    if (child_thread_idx >= max_num_threads) {
	oldmax = max_num_threads;
	max_num_threads += MAX_NUM_THREADS_INCR;
	new_xpos = new double[max_num_threads];
	new_ypos = new double[max_num_threads];
	new_position_index = new int[max_num_threads];
	for (i = 0; i < oldmax; i++) {
	    new_xpos[i] = xpos[i];
	    new_ypos[i] = ypos[i];
	    new_position_index[i] = position_index[i];
	}
	delete(xpos);
	delete(ypos);
	delete(position_index);
	xpos = new_xpos;
	ypos = new_ypos;
	position_index = new_position_index;
    }
    j = 0;
    while ((j < next_position) &&
	   (getThreadState(position_array[j]) != THREAD_TERMINATE)) j++;
    if (j == next_position) {
	if (next_position >= max_num_positions) {
	    oldmax = max_num_positions;
	    max_num_positions += MAX_NUM_THREADS_INCR;
	    new_position_array = new int[max_num_positions];
	    for (i = 0; i < oldmax; i++) {
		new_position_array[i] = position_array[i];
	    }
	    delete(position_array);
	    position_array = new_position_array;
	}
	next_position++;
    }

    position_array[j] = child_thread_idx;
    position_index[child_thread_idx] = j;
    xpos[child_thread_idx] = origx +
      (timestamp-init_timestamp) * screen_width_units_inv;
    ypos[child_thread_idx] = origy + yincr*position_index[child_thread_idx];
    line = new Line(polkaview, True,
		    xpos[thread_idx], ypos[thread_idx]+thickness,
		    0.0, ypos[child_thread_idx]-ypos[thread_idx],
		    color_depot->color(child_thread_idx));
    line->Originate(global_time);
}


void
ThreadLifeTime_polka::lifetimeJoinThread(int thread_idx,
					 unsigned long,  /* timestamp */
					 int join_thread_idx)
{
    Line  *line;

    /* DBG3("Join Thread: thread %d, timestamp %d, join thread %d\n",
	 thread_idx, timestamp, join_thread_idx); */
    line  = new Line(polkaview, True,
		     xpos[join_thread_idx], ypos[join_thread_idx],
		     0.0, ypos[thread_idx] - ypos[join_thread_idx],
		     color_depot->color(join_thread_idx));
    line->Originate(global_time);
}

void
ThreadLifeTime_polka::lifetimeChangeState(int thread_idx,
					  unsigned long timestamp,
					  ThreadStateType prev_state)
{
    Rectangle  *rect;
    char       *colname;
    double     fill;
    double     new_pos;

    colname = color_depot->color(thread_idx);
    if (prev_state == THREAD_RUNNING) {
	fill = 1.0;
    } else if (prev_state == THREAD_READY) {
	// colname = "#ffe7ba";   // wheat1
	fill = 0.40;
    } else if (prev_state == THREAD_WAITING_MUTEX) {
	// colname = "#c9c9c9";   // gray79
	colname = (char*)"black";
	fill = 0.2;
    } else if (prev_state == THREAD_WAITING_CONDITION) {
	// colname = "#7f7f7f";   // gray50
	colname = (char*)"black";
	fill = 0.5;
    } else if (prev_state == THREAD_WAITING_JOIN) {
	// colname = "#404040";   // gray25
	colname = (char*)"black";
	fill = 0.8;
    } else if (prev_state == THREAD_EXIT) {
	// colname = "red";  // wrong state
	fill = 0.1;
    } else if (prev_state == THREAD_PERTURBATION) {
#ifdef BLACK_WHITE
	colname = (char*)"black";
#else
	colname = (char*)"red";
#endif
	fill = 0.65;
    } else {
	DBG2("Thread %lu (%d) in error state\n", threadID(thread_idx),
	     thread_idx);
	colname = (char*)"red";
	fill = 1.0;
    }
    new_pos =  origx + (timestamp-init_timestamp) * screen_width_units_inv;
    if (new_pos > xpos[thread_idx]) {
	rect = new Rectangle(polkaview, True,
			     xpos[thread_idx], ypos[thread_idx],
			     new_pos - xpos[thread_idx], thickness,
			     colname, fill);
	rect->Originate(global_time);
	lifetimeHScroll(xpos[thread_idx], new_pos);
    }
    xpos[thread_idx] = new_pos;
}


void
ThreadLifeTime_polka::lifetimeHScroll(double begin_pt,
				      double end_pt)
{
    double  width;
    double  x1, y1, x2, y2;

    polkaview->GetCoord(x1, y1, x2, y2);
    if ((end_pt > x2) && (begin_pt < x2)) {
	width = x2 - x1;
	polkaview->SetCoord(end_pt - width * 0.25, y1,
			    end_pt + width * 0.75, y2);
	polkaview->start_ts = (end_pt-width*0.25)*screen_width;
	/* polkaview->start_ts += polkaview->screen_length*0.75; */
	polkaview->ClearTimeValueArea();
	/* polkaview->DrawTimeValues(); */
    }
}

void
tlresize(Widget, XtPointer thisone, XtPointer)
{
    TimedView *t = (TimedView *)thisone;
    t->DrawTimeLineTicks();
    t->DrawTimeValues();
    t->DrawTSLabel();
}

void
clrmapresize(Widget, XtPointer thisone, XtPointer)
{
    TimedView *t = (TimedView *)thisone;
    t->DrawLegend();
}


static void
panlCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,change;

    v->GetCoord(lx,by,rx,ty);
    change = (rx-lx) * hmove_factor;
    v->SetCoord(lx-change, by, rx-change, ty);
    v->start_ts -= v->screen_length * hmove_factor;
    v->DrawTimeValues();
}


static void
panrCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,change;

    v->GetCoord(lx,by,rx,ty);
    change = (rx-lx) * hmove_factor;
    v->SetCoord(lx+change, by, rx+change, ty);
    v->start_ts += v->screen_length*hmove_factor;
    v->DrawTimeValues();
}


static void
pandCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,change;

    v->GetCoord(lx,by,rx,ty);
    change = (ty-by) * vmove_factor;
    v->SetCoord(lx, by-change, rx, ty-change);
}


static void
panuCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,change;

    v->GetCoord(lx,by,rx,ty);
    change = (ty-by) * vmove_factor;
    v->SetCoord(lx, by+change, rx, ty+change);
}

static void
zoominCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,xc,yc,halfw,halfh;

    v->GetCoord(lx,by,rx,ty);
    xc = (lx+rx)*0.5;
    yc = (by+ty)*0.5;
    halfw = (rx-lx)*0.5*hzoom_factor;
    halfh = (ty-by)*0.5;
    v->SetCoord(xc-halfw, yc-halfh, xc+halfw, yc+halfh);
    v->start_ts += (1-hzoom_factor)*v->screen_length*0.5;
    v->screen_length *= hzoom_factor;
    v->DrawTimeValues();
    v->DrawTSLabel();
}

static void
zoomoutCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,xc,yc,halfw,halfh,zoomfactor = 1.0/hzoom_factor;

    v->GetCoord(lx,by,rx,ty);
    xc = (lx+rx)*0.5;
    yc = (by+ty)*0.5;
    halfw = (rx-lx)*0.5*zoomfactor;
    halfh = (ty-by)*0.5;
    v->SetCoord(xc-halfw, yc-halfh, xc+halfw, yc+halfh);
    v->start_ts += (1-zoomfactor)*v->screen_length*0.5;
    v->screen_length *= zoomfactor;
    v->DrawTimeValues();
    v->DrawTSLabel();
}

static void
vzoominCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,xc,halfw,halfh2;

    v->GetCoord(lx,by,rx,ty);
    xc = (lx+rx)*0.5;
    halfw = (rx-lx)*0.5;
    halfh2 = (ty-by)*vzoom_factor;
    v->SetCoord(xc-halfw, by, xc+halfw, by+halfh2);
}

static void
vzoomoutCB(Widget, XtPointer client_data, XtPointer)
{
    TimedView *v = (TimedView *)client_data;
    double lx,by,rx,ty,xc,halfw,halfh2,zoomfactor = 1.0/vzoom_factor;

    v->GetCoord(lx,by,rx,ty);
    xc = (lx+rx)*0.5;
    halfw = (rx-lx)*0.5;
    halfh2 = (ty-by)*zoomfactor;
    v->SetCoord(xc-halfw, by, xc+halfw, by+halfh2);
}

