/***************************************************************/
/*							       */
/*			     POLKA.H			       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#ifndef _POLKA_H
#define _POLKA_H


#define XTFUNCPROTO
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>


/****** Forward declarations to make some compilers happy **********/

class View;
class AnimObject;
class ActionInfo;
class ActionDeleteNode;
class ImplNode;
class IntNode;
class TimerCBNode;
class OFFSET;



/* ********************************************************************
Initialization
******************************************************************** */

typedef void (*APP_CALLBACK)(Widget, XtPointer, XtPointer);

void PolkaInitialize(Display *, XtAppContext, Widget);
void PolkaSetCallbacks(APP_CALLBACK, APP_CALLBACK, APP_CALLBACK,
                       APP_CALLBACK);
void PolkaReset();



/* ********************************************************************

Class View

******************************************************************** */

const int REFRESH_ALL      = 0;
const int REFRESH_MODIFIED = 1;


typedef int ResizeMode;

const ResizeMode ConstantAspect = 0;
const ResizeMode CoordStretch   = 1;


typedef  int   COLORINDEX;

typedef void (*TIMER_CALLBACK)(View *, int, void*);



typedef struct AONode
{
   class  AnimObject  *object;
   struct AONode      *next;
   struct AONode      *prev;
   struct AONode      *upnext;
   struct AONode      *upprev;
} *AOPtr;


class View {
 friend class AnimObject;    // maybe make these specific mem fcts later
 friend class AnimObjectImpl;
 friend void resize(Widget, XtPointer, XtPointer);
 friend void SelectPoint(Widget, XtPointer, XEvent *, Boolean *);

 private:

   void setupWindow(const char *);
   void animNextFrame();
   void clear();

   void originateAnimObjects(int);
   void updateAll(int);
   void drawAll();
   void drawAllUpdaters();
   void nextFrame();

   void fastDraw();
   void fastOriginate(int);

 protected:
   int     time;

   /*** COMMENT: a number of variables and procedure definitions are
        moved from private to protected, so that subclasses with more
	features can be derived from class View. ***/
   int          created;          /* has window been made for it yet?       */
   int          debug;            /* is debugging on?                       */
   int          redrawall;        /* redraw all objects for next frame?      */
   Widget       topshell;         /* top widget for View                    */
   Widget       easel;	          /* onscreen workSpace widget to 
                                                         display anim frames */
   Window	destwin; 	  /* current X window final target, blt into */
   Pixmap       window;           /* offscreen to draw into, blt from        */
   COLORINDEX   bgcolor;          /* X bg color for window                   */
   ResizeMode   resizeMode;       /* what to do when viewer resizes          */
   int		pixX;		  /* window's width			     */
   int		pixY;		  /*          height			     */
   double       aspectx;          /* aspect ratio, pixels -> coords          */
   double       aspecty;          /* aspect ratio, pixels -> coords          */
   double	olx,oby,orx,oty;  /* originals				     */
   double	lx,by,rx,ty;	  /* working vals			     */
   double       dx,dy;
   double       damlx,damrx,
                damby,damty;      /* damaged bounding box                    */
   double       boxpad;           /* padding to check object intersection    */
   int          needrefresh;      /* do next animation frame by refresh?     */
   int          getMouse;	  /* Waiting for mouse input (T/F)	     */
   int	        mousex,mousey;    /*    Coords of mouse input		     */
   int		numobjects;	  /* # of basic AnimObjects in window        */
   int		motionblur;	  /* disable clear for printing window dumps */
   AOPtr	AOHead;   	  /* current configuration of AnimObjects    */
   AOPtr	AOTail;
   AOPtr        PendingHead;      /* AnimObjects waiting to join the world   */
   AOPtr        PendingTail;
   AOPtr        UpdateHead;       /* AnimObjects with pending Actions        */
   AOPtr        UpdateTail;
   AnimObject  *updatingObject;   /* Used to determine where                 */
                                  /* attachment chain originated from        */
   TimerCBNode *timerCallbacks;   /* list of user callbacks at a frame #     */
   int          inTimerCallback;  /* can't call Animate if true              */

   APP_CALLBACK AppPanLeftCB, AppPanRightCB, AppPanUpCB, AppPanDownCB;
				  /* user supplied panning callbacks         */
   APP_CALLBACK AppZoomInCB, AppZoomOutCB;
                                  /* user supplied zooming callbacks         */

   void prepareAnim(Window);
   inline void fixPad() { boxpad = 2.0 / (double)pixX * dx; };
   inline double  ROUNDER(double Z) {return( Z > 0.0 ? 
                                        (Z + 0.5) : (Z - 0.5) );};

 public:
   View();
   ~View();
   void SetDebug(int d) { debug = d; };
   void SetRedrawAll(int r) { redrawall = r; }
   void RefreshNeeded(int r) { needrefresh = r; };
   void AlterLLCoord(double, double);
   void AlterURCoord(double, double);
   void SetCoord(double, double, double, double);
   void GetCoord(double& ilx, double& iby, double& irx, double& ity)
       { ilx = lx; iby = by; irx = rx; ity = ty; };
   void GetDimension(int& x, int& y)
       { x = pixX; y = pixY; };
   void SetBgColor(char *);
   void SetPanCallbacks(APP_CALLBACK, APP_CALLBACK, APP_CALLBACK,
			APP_CALLBACK, APP_CALLBACK, APP_CALLBACK);
   int Create(const char *title="Polka", ResizeMode rm=CoordStretch,
              int wid=0, int hei=0);
                // We want to take wid and hei defaults from resource file
   int Animate(int, int);
   int AnimateOne(int);
   int Simulate(int, int);
   int FastAnimate(int, int);
   void RegisterTimerCallback(int, TIMER_CALLBACK, void*);
   int CheckInput();

   int PickCoord(double&, double&);
   int PickAnimObject(class AnimObject* &);

   void Refresh();
   inline int     PIXX() { return(pixX); };
   inline int     PIXY() { return(pixY); };
   inline int     TO_PIX_X(double A)     
         { return( (int) ROUNDER( pixX * (((A)-lx)/dx) ) ); };
   inline int     TO_PIX_Y(double A)     
         { return( (int) ROUNDER( pixY * ((dy-(A)+by)/dy) ) ); };
   inline double  TO_REAL_X(int B)    
         { return( ((double)(B)/(double)pixX) * dx + lx ); };
   inline double  TO_REAL_Y(int B)    
         { return( ((double)(pixY-B)/(double)pixY) * dy + by ); };
   inline int     SIZE_PIX_X(double A)   
         { return( (int) ROUNDER( pixX * ((A)/dx)) ); };
   inline int     SIZE_PIX_Y(double A)   
         { return( (int) ROUNDER( pixY * ((A)/dy)) ); };
   inline double  SIZE_REAL_X(int B)  
         { return( ((double)(B)/(double)pixX) * dx); };
   inline double  SIZE_REAL_Y(int B)  
         { return( ((double)(B)/(double)pixY) * dy); };
   inline Window DrawWindow() { return window;};
   inline int GetDebug() { return debug; };
   inline COLORINDEX GetBgColor() { return bgcolor; };
   void DamageCheck(double lx1, double by1, double rx1, double ty1) {
      if (lx1 < damlx) damlx = lx1;
      if (rx1 > damrx) damrx = rx1;
      if (by1 < damby) damby = by1;
      if (ty1 > damty) damty = ty1;
      };
   AnimObject *GetUpdatingObject() { return updatingObject; };
   void AddUpdateObject(AOPtr);
   void RemoveUpdateObject(AOPtr);
};
      




/* ********************************************************************

Constituent Classes Loc, AnimObject, and Action

******************************************************************** */


/*** forward declarations ***/
typedef class Loc       *LocPtr;
typedef class Action    *ActionPtr;



/* ********************************************************************
Class Loc
******************************************************************** */

#ifndef _EQUAL
#define  _EQUAL(A,B)   ( ((A) - 0.000001 < (B)) && ((B) < (A) + 0.000001) )
#endif


class Loc {
   private:
      double	x;
      double	y;

   public:
      Loc() { x = 0.0; y = 0.0; };
      Loc(double x1, double y1) {
         this->x = x1;
         this->y = y1;
         };
      ~Loc() { };

      LocPtr Copy() { return ( new Loc (x,y) ); };
      LocPtr Modify(double x1, double y1) {
         return ( new Loc (this->x+x1,this->y+y1) );
         };
      void Alter(double newx, double newy) { x=newx; y=newy; };
      double XCoord() { return( x ); };
      double YCoord() { return( y ); };
      int operator == (Loc &loc) {
         return( _EQUAL(x,loc.x) && _EQUAL(y,loc.y) );
      };
};


class Loc1Desc {
   public:
      int num;
      double fx,fy;
      double tx,ty;
      Loc1Desc();
      int Make(Loc*[]);
} ;

class Loc2Desc {
   public:
      int rows;
      int cols;
      double llx,lly;
      double urx,ury;
      Loc2Desc();
      int Make(Loc*[]);
 } ;



/* ********************************************************************
Class AnimObject
******************************************************************** */

typedef  int   LINE_STYLE, FILL_STYLE;
typedef  char *COLOR;
typedef  int   PART;

#define  SOLID	        0
#define  DASHED         1
#define  DOTTED         2
#define  THICK	        3
#define  THICK_DASHED   4
#define  THICK_DOTTED   5
#define  THICKER	6
#define  THICKER_DASHED 7
#define  THICKER_DOTTED 8


#define  PART_C	      0
#define  PART_NW      1
#define  PART_N	      2
#define  PART_NE      3
#define  PART_E	      4
#define  PART_SE      5
#define  PART_S	      6
#define  PART_SW      7
#define  PART_W	      8


enum AnimObjectType {
   P_Line,
   P_Rectangle,
   P_Circle,
   P_Ellipse,
   P_Polygon,
   P_Polyline,
   P_Pie,
   P_Spline,
   P_Text,
   P_Bitmap,
   P_Set,
   P_Misc
};


typedef struct _AttachNode
{
   class  AnimObject *object;
   PART              dept;
   PART              indept;
   double            xoffset;
   double            yoffset;
   struct _AttachNode *next;
} AttachNode, *Attachment;




class AnimObjectImpl {
  private:
    virtual void transSpecial(char*, double, double) {};
  protected:
    View *view;
    class AnimObject *wrapper;
    AnimObjectType type;
    int visibility;
    double locx,locy;
  public:
    AnimObjectImpl(View *, int, double, double);
    AnimObjectImpl(const AnimObjectImpl&);
    virtual ~AnimObjectImpl() {};
    virtual AnimObjectImpl *clone() const = 0; // virtual constructor
    void SetWrapper(AnimObject *w) { wrapper = w; };
    void Warp(double x, double y) { locx=x; locy=y; }; // AnimObject::doChanges
    void MyLoc(double& x, double& y) { x=locx; y=locy; } // uses these two
    AnimObjectType Type() { return type; };
    View *MyView() { return view; };
    void DamageIt();
    virtual void Trans(char*, double, double);
    virtual void Draw() = 0;
    virtual void Erase() = 0;
    virtual LocPtr Where(PART) { return new Loc(locx,locy); };
    virtual void BoundBox(double *, double *, double *, double *) = 0;
};




class LineImpl: public AnimObjectImpl {
   private:
    double sizex,sizey;
    char colorname[32];
    COLORINDEX color;
    double width;
    double style;
    int arrow;
    double arrowf1x,arrowf1y;
    double arrowf2x,arrowf2y;
    double arrowf3x,arrowf3y;
    double arrowf4x,arrowf4y;
    void drawer(COLORINDEX);
    void transSpecial(char*, double, double);
  public:
    LineImpl(View *view, int v, double lx, double ly, double sx, double sy, 
	     COLOR c=(char*)"black", double w=0.0, double s=1.0, int a=0);
    LineImpl(const LineImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new LineImpl(*this); };
    ~LineImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int*, double*, double*, 
                   double*, double*, COLOR, double*,
                   double*, int*);
};


class RectangleImpl: public AnimObjectImpl {
  private:
    double sizex,sizey;
    char colorname[32];
    COLORINDEX color;
    double fill;
    void drawer(COLORINDEX, FILL_STYLE);
    void transSpecial(char*, double, double);
  public:
    RectangleImpl(View *view, int v, double lx, double ly, 
                 double sx, double sy, COLOR c=(char*)"black", double f=0.0) ;
    RectangleImpl(const RectangleImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new RectangleImpl(*this); };
    ~RectangleImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, 
                 double *, double *, COLOR, double *);
  };
    
class CircleImpl: public AnimObjectImpl {
  private:
    double radius;
    char colorname[32];
    COLORINDEX color;
    double fill;
    void drawer(COLORINDEX, FILL_STYLE);
    void transSpecial(char*, double, double);
  public:
    CircleImpl(View *view, int v, double lx, double ly, double r, 
	       COLOR c=(char*)"black", double f=0.0);
    CircleImpl(const CircleImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new CircleImpl(*this); };
    ~CircleImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, double *, 
              COLOR, double *);
  };
    

class EllipseImpl: public AnimObjectImpl {
  private:
    double sizex,sizey;
    char colorname[32];
    COLORINDEX color;
    double fill;
    void drawer(COLORINDEX, FILL_STYLE);
    void transSpecial(char*, double, double);
  public:
    EllipseImpl(View *view, int v, double lx, double ly, 
                double sx, double sy, COLOR c=(char*)"black", double f=0.0);
    EllipseImpl(const EllipseImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new EllipseImpl(*this); };
    ~EllipseImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, 
                double *, double *, COLOR, double *);
  };
    
class PolylineImpl: public AnimObjectImpl {
  private:
    int vertices;
    double vx[32];
    double vy[32];
    char colorname[32];
    COLORINDEX color;
    double width;
    double style;
    int arrow;
    double arrowf1x,arrowf1y;
    double arrowf2x,arrowf2y;
    double arrowf3x,arrowf3y;
    double arrowf4x,arrowf4y;
    void drawer(COLORINDEX);
    void transSpecial(char *, double, double);
  public:
    PolylineImpl(View *view, int v, double lx, double ly, int ve, 
                double vtx[], double vty[], COLOR c=(char*)"black", 
                double w=0.0, double s=1.0, int a=0);
    PolylineImpl(const PolylineImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new PolylineImpl(*this); };
    ~PolylineImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, int *, 
                double[], double[], COLOR, 
                double *, double *, int *);
  };

class PieImpl: public AnimObjectImpl {
  private:
    double radius;
    double begangle;
    double delta;
    char colorname[32];
    COLORINDEX color;
    double fill;
    void drawer(COLORINDEX, FILL_STYLE);
    void transSpecial(char*, double, double);
  public:
    PieImpl(View *view, int v, double lx, double ly, double r, 
             double ba=0.0, double del=1.0, COLOR c=(char*)"black", double f=0.0);
    PieImpl(const PieImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new PieImpl(*this); };
    ~PieImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, 
                double *, double *, double *, 
                COLOR, double *);
  };
    

class SplineImpl: public AnimObjectImpl {
  private:
    int vertices;
    double vx[32];
    double vy[32];
    char colorname[32];
    COLORINDEX color;
    double width;
    double style;
    void drawer(COLORINDEX);
    void transSpecial(char *, double, double);
  public:
    SplineImpl(View *view, int v, double lx, double ly, int ve, 
                double vx[], double vy[], COLOR c=(char*)"black", double w=0.0, 
                double s=1.0);
    SplineImpl(const SplineImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new SplineImpl(*this); };
    ~SplineImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, int *, 
                double[], double[], COLOR, double *, double *);
  };

class PolygonImpl: public AnimObjectImpl {
  private:
    int vertices;
    double vx[32];
    double vy[32];
    char colorname[32];
    COLORINDEX color;
    double fill;
    void drawer(COLORINDEX, FILL_STYLE);
    void transSpecial(char *, double, double);
  public:
    PolygonImpl(View *view, int v, double lx, double ly, int vert, 
                double vx[], double vy[], COLOR c=(char*)"black", double f=0.0);
    PolygonImpl(const PolygonImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new PolygonImpl(*this); };
    ~PolygonImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, int *, 
                double[], double[], COLOR, double *);
  };

class TextImpl: public AnimObjectImpl {
  private:
    char colorname[32];
    COLORINDEX color;
    char fontname[128];
    int fontid;
    char tstring[128];
    int orient;
    int xext,xoff;
    int yext,yoff;
    void drawer(COLORINDEX);
    void transSpecial(char*, double, double);
  public:
    TextImpl(View *view, int v, double lx, double ly, COLOR c, 
            const char *fn, const char *ts, int o);
    TextImpl(const TextImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new TextImpl(*this); };
    ~TextImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, COLOR, 
            char*, char*, int *);
};
    

class BitmapImpl: public AnimObjectImpl {
  private:
    Pixmap pixmap;
    int iwidth,iheight;
    char fgcolname[32];
    char bgcolname[32];
    char *bmdata;
    void transSpecial(char*, double, double);
  public:
    BitmapImpl(View *, int, double, double, int, int, char[], COLOR, COLOR);
    BitmapImpl(const BitmapImpl&);
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new BitmapImpl(*this); };
    ~BitmapImpl();
    LocPtr Where(PART);
    void BoundBox(double*, double*, double*, double*);
    void Draw();
    void Erase();
    void GetValues(View**, int *, double *, double *, 
            int *, int *, char[], COLOR, COLOR);
};
    

class SetImpl: public AnimObjectImpl {
  public: 
    SetImpl(View *v);
    SetImpl(const SetImpl& s) : AnimObjectImpl(s) {};
    AnimObjectImpl *clone() const     // redef of virtual constructor
       { return new SetImpl(*this); };
    LocPtr Where(PART);
    void BoundBox(double *, double *, double *, double *);
    void Draw();
    void Erase();
};





typedef  void (*CALLBACKFCT)(AnimObject*, void *, double, double);

class AnimObject {
  friend void SelectPoint(Widget, XtPointer, XEvent *, Boolean *);
  private:
    View *view;
    int originated;
    int originateAt;
    int modified;
    int updating;
    void *objData;
    void (*callbackFct)(AnimObject *, void *, double, double);
    void *callbackData;
    struct ActionInfo *first;       // list of regular Actions
    struct ActionInfo *last;
    struct ActionInfo *firstContinuous; // list of continuous running Actions
    struct ActionInfo *lastContinuous;
    struct ImplNode *changers;
    struct IntNode *deleters;
    struct ActionDeleteNode *continDeleters;
    Attachment attachments;
    AnimObject *attachedTo;
    AnimObject *attachDoneFrom;
    AOPtr viewAOP;
    void addObject();
    void deleteMe(); 
    int doDeletes(int);
    void doContDeletes(int);
    void doChanges(int);
    Attachment getAttachments() { return attachments; };
    void addAttachment(Attachment);
    void removeAttachment(AnimObject *);
    void propagateAttachment(AnimObject *, Attachment, AnimObject *);
    struct ActionInfo *DeleteActionInfoPtr(struct ActionInfo *);
    void deleteContinActionInfoPtr(ActionPtr);
  protected:
    AnimObjectImpl *object;
    AOPtr objs;  // things in set, or sets in which object (non-set) is in
    AnimObjectType type;
    AnimObject(AnimObjectImpl *p);
    AnimObject& operator=(const AnimObject&);
  public:
    AnimObject(const AnimObject&);
    ~AnimObject() 
       { deleteMe();
         delete object;
         object = 0; }
    int operator==(const AnimObject* a) 
       { return (this == a); }
    void Originate(int);
    void Change(int, AnimObject *, int useoldpos=0);
    void Delete(int);
    int Program(int, ActionPtr);
    int ProgramContinuous(int, ActionPtr);
    void RemoveContinuous(int, ActionPtr);
    LocPtr Where(PART p) { return(object->Where(p)); };
    Attachment AttachTo(AnimObject *, PART, PART, double x=0.0, double y=0.0);
    int Detach(Attachment);
    void SetAttachTraverseObject(AnimObject *a) { attachDoneFrom=a; };
    void StoreData(void *od)  { objData = od; };
    void *RetrieveData() { return objData; };
    void SetCallback(CALLBACKFCT cb, void *d=NULL) {
            callbackFct = cb;
            callbackData = d;
            };

    void Update(int);       //  from here on, should not be used by outsiders
    void Draw() { object->Draw(); };
    void Erase() { object->Erase(); };
    void BoundBox(double *lx, double *by, double *rx, double *ty) 
           { object->BoundBox(lx,by,rx,ty); };
    void DamageIt() { object->DamageIt(); };
    int Intercepts(double, double, double, double);
    void Nullify() { object = 0; }; // called only by Change
    int OriginateTime() { return originateAt; };
    void SetOriginated(int o) { originated = o; };
    int Originated() { return originated; };
    int GetUpdating() { return updating; };
    View *MyView() { return view; };
    AnimObjectType Type() { return type; };
    AnimObjectImpl *Object() { return object; };
    void SetModify(int m) { modified = m; };
    int Modified() { return modified; };
    void IsInSet(AnimObject *);
    void SetIsGone(AnimObject *);
};


class Line: public AnimObject {
  public:
    Line(View *view1, int v, double lx, double ly, double sx, double sy, 
            COLOR c=(char*)"black", double w=0.0, double s=1.0, int a=0)
      : AnimObject( new LineImpl(view1, v, lx, ly, sx, sy, c, w, s, a) ) {};
    Line& operator=(const Line&);
    void GetValues(View**view1, int* v, double* lx, double* ly, 
                   double* sx, double* sy, COLOR c, double* w,
                   double* s, int* a)
          { ((LineImpl*)object)->GetValues(view1,v,lx,ly,sx,sy,c,w,s,a); };
};


class Rectangle: public AnimObject {
  public:
    Rectangle(View *view1, int v, double lx, double ly, 
                 double sx, double sy, COLOR c=(char*)"black", double f=0.0) 
      : AnimObject( new RectangleImpl(view1, v, lx, ly, sx, sy, c, f) ) {};
    Rectangle& operator=(const Rectangle&);
    void GetValues(View**view1, int *v, double *lx, double *ly, 
                 double *sx, double *sy, COLOR c, double *f)
          { ((RectangleImpl*)object)->GetValues(view1,v,lx,ly,sx,sy,c,f); };
};
    
class Circle: public AnimObject {
  public:
    Circle(View *view1, int v, double lx, double ly, double r, 
              COLOR c=(char*)"black", double f=0.0)
      : AnimObject( new CircleImpl(view1, v, lx, ly, r, c, f) ) {};
    Circle& operator=(const Circle&);
    void GetValues(View**view1, int *v, double *lx, double *ly, double *r, 
              COLOR c, double *f)
          { ((CircleImpl*)object)->GetValues(view1,v,lx,ly,r,c,f); };
};
    

class Ellipse: public AnimObject {
  public:
    Ellipse(View *view1, int v, double lx, double ly, 
                double sx, double sy, COLOR c=(char*)"black", double f=0.0)
      : AnimObject( new EllipseImpl(view1, v, lx, ly, sx, sy, c, f) ) {};
    Ellipse& operator=(const Ellipse&);
    void GetValues(View**view1, int *v, double *lx, double *ly, 
                double *sx, double *sy, COLOR c, double *f)
          { ((EllipseImpl*)object)->GetValues(view1,v,lx,ly,sx,sy,c,f); };
};
    
class Polyline: public AnimObject {
  public:
    Polyline(View *view1, int v, double lx, double ly, int ve, 
                double vtx[], double vty[], COLOR c=(char*)"black", 
                double w=0.0, double s=1.0, int a=0)
    : AnimObject( new PolylineImpl(view1, v, lx, ly, ve, vtx, vty, c, w, s, a))
          {};
    Polyline& operator=(const Polyline&);
    void GetValues(View**view1, int *v, double *lx, double *ly, int *ve, 
                double vtx[], double vty[], COLOR c, 
                double *w, double *s, int *a)
          { ((PolylineImpl*)object)->GetValues(view1,v,lx,ly,ve,vtx,vty,c,
                               w,s,a); };
};

class Pie: public AnimObject {
  public:
    Pie(View *view1, int v, double lx, double ly, 
                double rad, double bang=0.0, double del=1.0, 
                COLOR c=(char*)"black", double f=1.0)
      : AnimObject( new PieImpl(view1, v, lx, ly, rad, bang, del, c, f) ) {};
    Pie& operator=(const Pie&);
    void GetValues(View**view1, int *v, double *lx, double *ly, 
                double *rad, double *bang, double *del, 
                COLOR c, double *f)
          { ((PieImpl*)object)->GetValues(view1,v,lx,ly,rad,bang,del,c,f); };
};
    
class Spline: public AnimObject {
  public:
    Spline(View *view1, int v, double lx, double ly, int ve, 
                double vx[], double vy[], COLOR c=(char*)"black", double w=0.0, 
                double s=1.0)
      : AnimObject( new SplineImpl(view1, v, lx, ly, ve, vx, vy, c, w, s) ) {};
    Spline& operator=(const Spline&);
    void GetValues(View**view1, int *v, double *lx, double *ly, int *ve, 
                double vx[], double vy[], COLOR c, double *w, double *s)
          { ((SplineImpl*)object)->GetValues(view1,v,lx,ly,ve,vx,vy,c,w,s); };
};

class Polygon: public AnimObject {
  public:
    Polygon(View *view1, int v, double lx, double ly, int vert, 
                double vx[], double vy[], COLOR c=(char*)"black", double f=0.0)
      : AnimObject( new PolygonImpl(view1, v, lx, ly, vert, vx, vy, c, f) ) {} ;
    Polygon& operator=(const Polygon&);
    void GetValues(View**view1, int *v, double *lx, double *ly, int *vert, 
                double vx[], double vy[], COLOR c, double *f)
          { ((PolygonImpl*)object)->GetValues(view1,v,lx,ly,vert,vx,vy,c,f); };
};

class Text: public AnimObject {
  public:
    Text(View *view1, int v, double lx, double ly, COLOR c, 
	 const char *fn, const char *ts, int o)
      : AnimObject( new TextImpl(view1, v, lx, ly, c, fn, ts, o) ) {};
    Text& operator=(const Text&);
    void GetValues(View**view1, int *v, double *lx, double *ly, COLOR c, 
            char* fn, char* ts, int *o)
          { ((TextImpl*)object)->GetValues(view1,v,lx,ly,c,fn,ts,o); };
};
    

class Bitmap: public AnimObject {
  public:
    Bitmap(View *view1, int v, double lx, double ly, int wid, int hei, 
            char data[], COLOR fg, COLOR bg)
    : AnimObject( new BitmapImpl(view1, v, lx, ly, wid, hei, data, fg, bg) ) {};
    Bitmap& operator=(const Bitmap&);
    void GetValues(View**view1, int *v, double *lx, double *ly, 
            int *wid, int *hei, char data[], COLOR fg, COLOR bg)
        { ((BitmapImpl*)object)->GetValues(view1,v,lx,ly,wid,hei,data,fg,bg); };
};
    

class Set: public AnimObject {
  public: 
    Set(View * view1, int num, AnimObject *ao[]); 
    Set(const Set& s);
    ~Set();
    AOPtr Objs() { return objs; };
    void RemoveFromSet(AnimObject *);
  private:
    void another(AnimObject *);
    Set& operator=(const Set&);
};








enum Align {
   AlignBottom, AlignTop, AlignMiddle, AlignLeft, AlignRight
};

enum Spacing {
   SpacingNone, SpacingSame, SpacingAbsolute
};

class AnimObjectGroup {
   public:
     int horiz;
     Align align;
     int vis;
     char color[32];
     Spacing spacetype;
     double spacing;

     int useints;
     int *intvals;
     int intmax;
     int intmin;

     int usedoubles;
     double *doublevals;
     double doublemax;
     double doublemin;

     AnimObjectGroup();
};


class LineGroup : public AnimObjectGroup {
   public:
      double linewidth;
      double linestyle;
      int arrow;
      LineGroup();
      int Make(View*, Line *[], int, double, double, double, double);
};

class RectangleGroup : public AnimObjectGroup {
   public:
     double fill;
     RectangleGroup();
     int Make(View*, Rectangle *[], int, double, double, double, double);
};


class CircleGroup : public AnimObjectGroup {
   public:
     double fill;
     CircleGroup();
     int Make(View*, Circle *[], int, double, double, double, double);
};





/*********************************************************************
Class Action
******************************************************************** */


enum MOTION {
   STRAIGHT,
   CLOCKWISE,
   COUNTERCLOCKWISE
};



class Action {
   friend AnimObject;

   private:
      char           type[32];
      struct OFFSET *head;
      struct OFFSET *tail;
      int  	     count;
      void               AddOffsetAtEnd(double x, double y);
      void               DeleteOffsetFromFront();
      void               DeleteOffsetFromEnd();

   public:
      Action() { *type= '\0'; head = tail = NULL; count = 0; };
      Action(char *t) { strcpy(type,t); head = tail = NULL; count = 0; };
      Action(char *, int, double [], double []);
      Action(char *, int);
      Action(char *, MOTION);
      Action(char *, COLOR);
      Action(char *, LocPtr, LocPtr, MOTION);
      Action(char *, LocPtr, LocPtr, double);
      Action(char *, LocPtr, LocPtr, int);
      ~Action();

      void      Print();
inline int       Length() { return( count ); };
      double    Deltax();
      double    Deltay();
      ActionPtr AddHead(int);
      ActionPtr AddTail(int);
      ActionPtr DeleteHead(int);
      ActionPtr DeleteTail(int);
      ActionPtr Copy();
      ActionPtr ChangeType(char *);
      ActionPtr Reverse();
      ActionPtr Smooth();
      ActionPtr Rotate(int);
      ActionPtr Scale(double, double);
      ActionPtr Extend(double, double);
      ActionPtr Interpolate(double);
      ActionPtr Example(LocPtr, LocPtr);
      ActionPtr Iterate(int);
      ActionPtr Concatenate(ActionPtr);
      ActionPtr Compose(ActionPtr);
};



#endif    /* _POLKA_H */




