/***************************************************************/
/*							       */
/*			     POLKAWIDGETMOTIF.C		       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"
#include "polka_version.H"
#include <sys/types.h>	// For micro_sleep() 
#include <sys/time.h>	// For micro_sleep() 
#include <sys/socket.h>	// For micro_sleep() 

#include <StringDefs.h>
#include <Intrinsic.h>
#include <Shell.h>
#include <cursorfont.h>
#include <StringDefs.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/ArrowB.h>
#include <Xm/ScrollBar.h>




#define TITLE			"POLKA"
#define TITLEFONT		"variable"
#define TITLECOLOR		"Yellow"
#define BORDER_WIDTH		1
#define BDR_OFFSET		-BORDER_WIDTH

static void speedCB(Widget, XtPointer, XtPointer);
void debugCB(Widget, XtPointer, XtPointer);
void refreshCB(Widget, XtPointer, XtPointer);
void pauseCB(Widget, XtPointer, XtPointer);
void panlCB(Widget, XtPointer, XtPointer);
void panrCB(Widget, XtPointer, XtPointer);
void pandCB(Widget, XtPointer, XtPointer);
void panuCB(Widget, XtPointer, XtPointer);
void zoominCB(Widget, XtPointer, XtPointer);
void zoomoutCB(Widget, XtPointer, XtPointer);
void micro_sleep(int);

int           _started = 0;
Display      *_display = NULL;
XtAppContext  _context = NULL;
int           _color_screen;



void
PolkaInitialize(Display *d, XtAppContext c)
{
   _display = d;
   _context = c;
}









/*---------------------------------------------------------------------------*/
/*                              POLKA Widgets                                */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*                             toplevel                                      */
/*                                |                                          */
/*                               form                                        */
/*       /---------------/--------+---\-------\-----\-----\------\-----\     */
/*     title           rcleft         sbar   pause mode debug refresh quit   */
/*               /----/---\---\---\--\                                       */
/*            left right up down in out        			             */
/*                                                                           */
/*                                                                           */
/*	modeMgrShell			debugMgrShell		      */
/*	      |				      |			      */
/*	modeMenuMgr			debugMenuMgr		      */
/*	      |				      |			      */
/*	 modeShell			 debugShell		      */
/*	      |				      |			      */
/*	  Refresh			    Debug		      */
/*    /---------|-------\		    /-+-\		      */
/* by frame  by scene  none		  *Off  On		      */
/*---------------------------------------------------------------------------*/

/**************************************************************/
/* TANGO_layout_window -- Create, init (hardcoded), place, &  */
/*			  realize all widgets.		      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
View::setupWindow(const char *title)
{
   Widget      form, botform, botrightrowcol, runbut, debugbut,
                refreshbut, quitbut,
                botleftrowcol, larrowbut, rarrowbut, uarrowbut, darrowbut, 
                zoominbut, zoomoutbut, 
                leftform, fastlabel, speedbar, slowlabel;
   int         scr;
   XFontStruct *xfs;
   XmFontList fontlist;
   XmString    xmstr;
   XColor      color,exact;
   Arg	       args[20];
   int         n;
   static int  count = 2;
   /* Fake command-line arguments for XtInitialize() */
   unsigned int argc = 1;
   char *argv[1];

   if (debug)
      printf("DEBUG: View::setupWindow\n");

   if (!_context) {
      if (title) {
         argv[0] = new char[strlen(title)+1];
         strcpy(argv[0],title);
         }
      else {
         argv[0] = new char[strlen("polka")+1];
         strcpy(argv[0],"polka");
         }

      XtToolkitInitialize();
      _context = XtCreateApplicationContext();
      XtDisplayInitialize(_context, _display, title, "Polka",
             NULL, 0, &argc, argv);
   }
   topshell = XtAppCreateShell(title, "Polka", applicationShellWidgetClass,   
                        _display, NULL, 0);

   scr = DefaultScreen(_display);
   xfs = XLoadQueryFont(_display, "variable");
   fontlist =  XmFontListCreate(xfs, XmSTRING_DEFAULT_CHARSET);
   n = 0;
   XtSetValues(topshell, args, n);

   form = XtCreateManagedWidget("topform", xmFormWidgetClass, topshell, NULL, 0);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
   botform = XtCreateManagedWidget("botform", xmFormWidgetClass, form,
                                     args, n);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
/*
      XtSetArg(args[n], XmNnumColumns, 2); n++;
      XtSetArg(args[n], XmNpacking, XmPACK_COLUMN); n++;
*/
   botleftrowcol = XtCreateManagedWidget("botleftrowcol", 
                         xmRowColumnWidgetClass, botform, args, n);

      n = 0;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
   botrightrowcol = XtCreateManagedWidget("botrightrowcol", 
                       xmRowColumnWidgetClass, botform, args, n);

//      n = 0;
//      XtSetArg(args[n], XmNfontList,    fontlist); n++;
//      xmstr = XmStringCreate("Run animation",XmSTRING_DEFAULT_CHARSET); 
//      XtSetArg(args[n], XmNlabelString, xmstr); n++;
//   runbut = XtCreateManagedWidget("runbut", xmPushButtonWidgetClass,
//                      botrightrowcol, args, n);
//      XtAddCallback(runbut, XmNactivateCallback, pauseCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Debug",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
      XtSetArg(args[n], XmNindicatorOn, False); n++;
      XtSetArg(args[n], XmNshadowThickness, 3); n++;
   debugbut = XtCreateManagedWidget("debugbut", xmToggleButtonWidgetClass, 
                             botrightrowcol, args, n);
      XtAddCallback(debugbut, XmNvalueChangedCallback, debugCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Refresh",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   refreshbut = XtCreateManagedWidget("refreshbut", xmPushButtonWidgetClass, 
                             botrightrowcol, args, n);
      XtAddCallback(refreshbut, XmNactivateCallback, refreshCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Quit",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   quitbut = XtCreateManagedWidget("quitbut", xmPushButtonWidgetClass,
                            botrightrowcol, args, n);
      XtAddCallback(quitbut, XmNactivateCallback, quitCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_LEFT); n++;
   larrowbut = XtCreateManagedWidget("larrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(larrowbut, XmNactivateCallback, panlCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_RIGHT); n++;
   rarrowbut = XtCreateManagedWidget("rarrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(rarrowbut, XmNactivateCallback, panrCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_DOWN); n++;
   darrowbut = XtCreateManagedWidget("darrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(darrowbut, XmNactivateCallback, pandCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_UP); n++;
   uarrowbut = XtCreateManagedWidget("uarrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(uarrowbut, XmNactivateCallback, panuCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("In",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   zoominbut = XtCreateManagedWidget("zoominbut", xmPushButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(zoominbut, XmNactivateCallback, zoominCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Out",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   zoomoutbut = XtCreateManagedWidget("zoomoutbut", xmPushButtonWidgetClass, 
                                 botleftrowcol, args, n);
      XtAddCallback(zoomoutbut, XmNactivateCallback, zoomoutCB, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNbottomWidget, botform); n++;
   leftform = XtCreateManagedWidget("leftform", xmFormWidgetClass, 
                   form, args, n);

      n = 0;
      XtSetArg(args[n], XmNalignment,   XmALIGNMENT_CENTER); n++;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Fast",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
   fastlabel = XtCreateManagedWidget("fastlabel", xmLabelWidgetClass, 
                        leftform, args, n);

      n = 0;
      XtSetArg(args[n], XmNalignment,   XmALIGNMENT_CENTER); n++;
      XtSetArg(args[n], XmNfontList,    fontlist); n++;
      xmstr = XmStringCreate("Slow",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
   slowlabel = XtCreateManagedWidget("slowlabel", xmLabelWidgetClass, 
                        leftform, args, n);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNbottomWidget, slowlabel); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNtopWidget, fastlabel); n++;
      XtSetArg(args[n], XmNborderWidth, 1); n++;
   speedbar = XtCreateManagedWidget("speedbar", xmScrollBarWidgetClass, 
                     leftform, args, n);
      XtAddCallback(speedbar, XmNvalueChangedCallback, speedCB, 
                           (XtPointer)this);



      n = 0;
      XtSetArg(args[n], XtNborderWidth, BORDER_WIDTH); n++;
      XtSetArg(args[n], XmNforeground,  BlackPixel(_display,scr)); n++;
      XtSetArg(args[n], XmNbackground,  WhitePixel(_display,scr)); n++;

      XtSetArg(args[n], XmNresizePolicy, XmRESIZE_ANY); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNbottomWidget, botform); n++;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNleftWidget, leftform); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
   easel = XtCreateManagedWidget("easel", xmDrawingAreaWidgetClass, form,
				 args, n);
      XtAddCallback(easel, XmNexposeCallback, resize, (XtPointer)this);
      XtAddCallback(easel, XmNresizeCallback, resize, (XtPointer)this);
      XtAddEventHandler(easel, ButtonPressMask, FALSE, SelectPoint, NULL);

   XtRealizeWidget(topshell);
}



/**************************************************************/
/* TANGO_check_X_events -- Handle all pending X events.       */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
View::checkXEvents()
{
   XEvent xev;

   while (XtAppPending(_context) || paused) {
      XtAppNextEvent(_context, &xev);
      XtDispatchEvent(&xev);
      } 
}




/**************************************************************/
/* refresh -- (callback function) Redraw all images onscreen. */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
refreshCB(Widget, XtPointer thisone, XtPointer)
{
   ((View *)thisone)->Refresh();
}



/**************************************************************/
/* resize -- Resize widgets.				      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
resize(Widget w, XtPointer thisone, XtPointer)
{
   View *t = (View *)thisone;
   if (t->window) {	/* Avoid resize when widget initially displayed */
		/* -- *before* offscreen pixmap has been created.	*/
      t->prepareAnim(XtWindow(w));
      t->fixPad();
      t->Refresh();
      }
}



/***************************************************************/
/* delayset_callback -- Adjust the delay setting according     */
/*	       to the scrollbar position.		       */
/*							       */
/* RETURNS:  None.					       */
/***************************************************************/
void
speedCB(Widget, XtPointer thisone, XtPointer sl)
{
   XmScrollBarCallbackStruct *slider = (XmScrollBarCallbackStruct *)sl;
   View *v = (View *)thisone;
   v->SetDelay(5000 * slider->value);
}



/**************************************************************/
/* pause_callback -- Pause/unpause animation and change	      */
/*		     button label to reflect current state.   */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
pauseCB(Widget w, XtPointer client_data, XtPointer)
{
   Arg args[1];
   XmString newLabel;
   View *v;
   int paused;

   v = (View *)client_data;
   paused = v->GetPause();
   v->SetPause(paused ? 0 : 1);   // Initially paused as "run animation" 
   newLabel = XmStringCreate((!paused ? "unpause" : "pause"),
			     XmSTRING_DEFAULT_CHARSET);
   XtSetArg(args[0], XmNlabelString, newLabel);
   XtSetValues(w, args, 1);
   XmStringFree(newLabel);
}



/**************************************************************/
/* changeDebug_callback -- Change debug mode and set	      */
/*			   checkmark.			      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
debugCB(Widget, XtPointer client_data, XtPointer)
{
   View *v;
   int d;

   v = (View *)client_data;
   d = v->GetDebug();
   v->SetDebug( d ? 0 : 1);
}






void
panlCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (rx-lx) * panfactor;
   v->SetCoord(lx-change, by, rx-change, ty);
}


void
panrCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (rx-lx) * panfactor;
   v->SetCoord(lx+change, by, rx+change, ty);
}


void
pandCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (ty-by) * panfactor;
   v->SetCoord(lx, by-change, rx, ty-change);
}


void
panuCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (ty-by) * panfactor;
   v->SetCoord(lx, by+change, rx, ty+change);
}



void
zoominCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,changex,changey,zoomfactor = 0.8;

   v->GetCoord(lx,by,rx,ty);
   changey = ((ty-by) * (1.0-zoomfactor)) / 2.0;
   changex = ((rx-lx) * (1.0-zoomfactor)) / 2.0;
   v->SetCoord(lx+changex, by+changey, rx-changex, ty-changey);
}




void
zoomoutCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,changex,changey,zoomfactor = 0.8;

   v->GetCoord(lx,by,rx,ty);
   changey = ((ty-by) * (1.0/zoomfactor -1.0)) / 2.0;
   changex = ((rx-lx) * (1.0/zoomfactor -1.0)) / 2.0;
   v->SetCoord(lx-changex, by-changey, rx+changex, ty+changey);
}




/**************************************************************/
/* changeDebug_callback -- Change debug mode and set	      */
/*			   checkmark.			      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
quitCB(Widget, XtPointer thisone, XtPointer)
{
   View *v = (View *)thisone;
   XtDestroyWidget(v->topshell);
}


