
/* #include <stdio.h> */
#include "dv_cminterf.h"
#include "tuxian.H"
#include "general.h"

extern int debug_flag;


/**********************************************************************
 *   metheads of class Tuxian
 **********************************************************************/

Tuxian::Tuxian()
{
    init_timestamp = 0;
    timestamp_resolution = 1000000;
    timestamp_resolution_inv = 1.0 / timestamp_resolution;

    num_threads = 0;
    max_num_threads = MAX_NUM_THREADS_INIT;
    thread_ids = new(unsigned long[max_num_threads]);
}

int
Tuxian::threadIndex(unsigned long  threadid)
{
    int  i = num_threads-1;
    while (i >= 0 && thread_ids[i] != threadid) i--;
    return i;
}

int
Tuxian::storeThreadID(unsigned long  thread_id)
{
    unsigned long  *thrds;
    if (num_threads >= max_num_threads) {
	max_num_threads += MAX_NUM_THREADS_INCR;
	thrds = new(unsigned long[max_num_threads]);
	for (int i = 0; i < num_threads; i++) {
	    thrds[i] = thread_ids[i];
	}
	delete(thread_ids);
	thread_ids = thrds;
    }
    thread_ids[num_threads] = thread_id;
    num_threads++;
    return (num_threads-1);
}


/**********************************************************************
 *   metheads of class ThreadMap
 **********************************************************************/

void
ThreadMap::create(char* data)
{
    DM_Type_thread_initPtr event = (DM_Type_thread_initPtr) data;

    if (data != NULL) {
	init_timestamp = event->header.timestamp;
	timestamp_resolution = event->timestamp_resolution;
	timestamp_resolution_inv = 1.0 / timestamp_resolution;
    }
    threadmapInitialize();
}


void
ThreadMap::update(char* data)
{
    int			     thisthrd, thread_idx;
    SimpleTraceDataRecordPtr event = (SimpleTraceDataRecordPtr) data;

    thisthrd = threadIndex(event->thread);
    if ((event->type != DM_MONITOR_EXIT) && (event->type != DM_THREAD_FORK)
	&& (thisthrd < 0)) {
	DBG3("thrdmap Event: type %d, timestamp %u, thread %lx: may be out of order\n",
	     event->type, event->timestamp, event->thread);
	return;
    }
    switch(event->type) {
      case DM_THREAD_FORK :
	thread_idx =
	  storeThreadID(((DM_Type_thread_forkPtr) data)->child_thread);
	threadmapNewThread(thread_idx,
			   ((DM_Type_thread_forkPtr) data)->child_thread,
			   (long unsigned int) event->timestamp,
			   ((DM_Type_thread_forkPtr) data)->on_processor,
			   ((DM_Type_thread_forkPtr) data)->name);
	break;
      case DM_THREAD_EXIT:
	threadmapThreadExit(thisthrd, event->thread, event->timestamp);
	break;
      case DM_THREAD_SET_NAME:
	thread_idx =
	  threadIndex(((DM_Type_thread_set_namePtr) data)->the_thread);
	if (thread_idx >= 0) {
	    threadmapNewName(thread_idx,
			     ((DM_Type_thread_set_namePtr) data)->name);
	}
	break;
      case DM_MONITOR_EXIT:
	threadmapThreadExit(0, event->thread, event->timestamp);
	break;
      default:
	break;
    }
}



/**********************************************************************
 *   metheads of class ThreadLifeTime
 **********************************************************************/

ThreadLifeTime::ThreadLifeTime(int tlr)
	:Tuxian()
{
    max_num_threads = MAX_NUM_THREADS_INIT;
    ts_tolerance = tlr;
    parent_thrd_idx = new(int[max_num_threads]);
    prev_timestamp = new(unsigned long[max_num_threads]);
    thread_state = new ThreadStateType[max_num_threads];
    thread_prev_state = new ThreadStateType[max_num_threads];
    thread_detach_flag = new char[max_num_threads];
}


ThreadLifeTime::~ThreadLifeTime()
{
    delete(parent_thrd_idx);
    delete(prev_timestamp);
    delete(thread_state);
    delete(thread_prev_state);
    delete(thread_detach_flag);
}

void
ThreadLifeTime::create(char* data)
{
    DM_Type_thread_initPtr event = (DM_Type_thread_initPtr) data;
    /* int                    thread_idx; */

    if (data != NULL) {
	init_timestamp = event->header.timestamp;
	timestamp_resolution = event->timestamp_resolution;
	timestamp_resolution_inv = 1.0 / timestamp_resolution;
    }
/*     thread_idx = storeThreadID(event->header.thread);
    parent_thrd_idx[thread_idx] = -1;
    prev_timestamp[thread_idx] = event->header.timestamp;
    thread_prev_state[thread_idx] = thread_state[thread_idx]
      = THREAD_RUNNING;
    thread_detach_flag[thread_idx] = False; */
    lifetimeInitialize();
/*    lifetimeNewThread(-1, event->header.timestamp, thread_idx,
		      "initial thread"); */
}


void
ThreadLifeTime::update(char* data)
{
    int              i, thisthrd, tmpthrd, thread_idx, oldmax;
    unsigned long    *pts, prevts = 0;
    int              *parentthrd;
    ThreadStateType  *thrdstate, *prevthrdstate;
    char             *thrddetachflag;
    ThreadStateType  prevstate = THREAD_NO_STATE;
    SimpleTraceDataRecordPtr event = (SimpleTraceDataRecordPtr) data;
    thisthrd = threadIndex(event->thread);
    if ((event->type != DM_MONITOR_EXIT) && (event->type != DM_THREAD_OFF_RUNQ)
	&& (event->type != DM_THREAD_FORK) && (thisthrd < 0)) {
	DBG3("Event: type %d, timestamp %lu, thread %lx: may be out of order\n",
	     event->type, event->timestamp, event->thread);
	return;
    }
    if (thisthrd >= 0) {
	prevstate = thread_state[thisthrd];
	prevts = prev_timestamp[thisthrd];
    }
    switch(event->type) {
      case SEVENT_PERTURBATION:
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	lifetimeChangeState(thisthrd, event->timestamp+event->perturbation,
			    THREAD_PERTURBATION);
	break;
      case DM_THREAD_FORK :
	thread_idx =
	  storeThreadID(((DM_Type_thread_forkPtr) data)->child_thread);
	if (thread_idx >= max_num_threads) {
	    oldmax = max_num_threads;
	    max_num_threads += MAX_NUM_THREADS_INCR;
	    parentthrd = new(int[max_num_threads]);
	    pts = new(unsigned long[max_num_threads]);
	    thrdstate = new(ThreadStateType[max_num_threads]);
	    prevthrdstate = new ThreadStateType[max_num_threads];
	    thrddetachflag = new char[max_num_threads];
	    for (i = 0; i < oldmax; i++) {
		parentthrd[i] = parent_thrd_idx[i];
		pts[i] = prev_timestamp[i];
		thrdstate[i] = thread_state[i];
		prevthrdstate[i] = thread_prev_state[i];
		thrddetachflag[i] = thread_detach_flag[i];
	    }
	    delete(parent_thrd_idx);
	    delete(prev_timestamp);
	    delete(thread_state);
	    delete(thread_prev_state);
	    delete(thread_detach_flag);
	    parent_thrd_idx = parentthrd;
	    prev_timestamp = pts;
	    thread_state = thrdstate;
	    thread_prev_state = prevthrdstate;
	    thread_detach_flag = thrddetachflag;
	}
	prev_timestamp[thread_idx] = event->timestamp;
	thread_state[thread_idx] = thread_prev_state[thread_idx]
	  = THREAD_READY;
	thread_detach_flag[thread_idx] = False;
	if (thisthrd >= 0) {
	    lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	    lifetimeNewThread(thisthrd, event->timestamp, thread_idx,
			      ((DM_Type_thread_forkPtr) data)->name);
	    prev_timestamp[thisthrd] = event->timestamp;
	    thread_prev_state[thisthrd] = thread_state[thisthrd];
	    thread_state[thisthrd] = THREAD_RUNNING;
	} else {
	    prev_timestamp[thread_idx] = event->timestamp;
	    parent_thrd_idx[thread_idx] = -1;
	    thread_state[thisthrd] = thread_prev_state[thread_idx]
	      = THREAD_RUNNING;
	    lifetimeNewThread(-1, event->timestamp, thread_idx,
			      ((DM_Type_thread_forkPtr) data)->name);
	}
	break;
      case DM_THREAD_BEGIN_JOIN:
	tmpthrd = threadIndex(((DM_Type_thread_begin_joinPtr)event)->join_to);
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_WAITING_JOIN;
	break;
      case DM_THREAD_END_JOIN:
	tmpthrd = threadIndex(((DM_Type_thread_end_joinPtr)event)->join_to);
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	lifetimeChangeState(tmpthrd, event->timestamp, thread_state[tmpthrd]);
	lifetimeJoinThread(thisthrd, event->timestamp, tmpthrd);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_RUNNING;
	thread_state[tmpthrd] = THREAD_TERMINATE;
	break;
      case DM_THREAD_DETACH:
	tmpthrd=threadIndex(((DM_Type_thread_detachPtr)event)->detach_from);
	if (tmpthrd < 0) {
	    fprintf(stderr, "Warning: Unknown thread %lx detached from %lx\n",
		    ((DM_Type_thread_detachPtr)event)->detach_from,
		    event->thread);
	} else {
	    thread_detach_flag[tmpthrd] = True;
	}
	break;
      case DM_MUTEX_BEGIN_LOCK:
	/* Mutex-begin-lock should be followed by eitherr mutex-end-lock
	   event or thread-off-runq event.  The former case means the thread
	   obtains the lock immediately, and the thread is running all the
	   time.  The later case means the thread IS blocked, and is brought
	   back in after it gets the lock.
	   */
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_WAITING_MUTEX;
	break;
      case DM_MUTEX_END_LOCK:
	if ((prevstate == THREAD_WAITING_MUTEX) &&
	    (event->timestamp - prevts <= ts_tolerance)) {
	    thread_state[thisthrd] = THREAD_RUNNING;
	    break;
	} else {
	    lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	    prev_timestamp[thisthrd] = event->timestamp;
	    thread_prev_state[thisthrd] = thread_state[thisthrd];
	    thread_state[thisthrd] = THREAD_RUNNING;
	    break;
	}
      case DM_THREAD_EXIT:
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_EXIT;
	if (thread_detach_flag[thisthrd]) {
	    thread_state[thisthrd] = THREAD_TERMINATE;
	}
	break;
      case DM_CONDITION_BEGIN_WAIT:
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_WAITING_CONDITION;
	break;
      case DM_CONDITION_END_WAIT:
	lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	prev_timestamp[thisthrd] = event->timestamp;
	thread_prev_state[thisthrd] = thread_state[thisthrd];
	thread_state[thisthrd] = THREAD_RUNNING;
	break;
      case DM_THREAD_OFF_RUNQ:
	if ((thisthrd >= 0) && (thread_state[thisthrd] != THREAD_EXIT)
	    && (thread_state[thisthrd] != THREAD_TERMINATE)) {
	    lifetimeChangeState(thisthrd, event->timestamp, prevstate);
	    prev_timestamp[thisthrd] = event->timestamp;
	    thread_prev_state[thisthrd] = thread_state[thisthrd];
	    if ((prevstate != THREAD_WAITING_MUTEX) &&
		(prevstate != THREAD_WAITING_CONDITION) &&
		(prevstate != THREAD_WAITING_JOIN) &&
		(prevstate != THREAD_EXIT) &&
		(prevstate != THREAD_TERMINATE)) {
		thread_state[thisthrd] = THREAD_RUNNING;
	    }
	}
	tmpthrd =
	  threadIndex(((DM_Type_thread_off_runqPtr) event)->thread_off_runq);
	if ((tmpthrd < 0) || (thread_state[tmpthrd] == THREAD_EXIT)
	    || (thread_state[tmpthrd] == THREAD_TERMINATE)) {
	    break;
	}
	prevstate = thread_state[tmpthrd];
	lifetimeChangeState(tmpthrd, event->timestamp, prevstate);
	prev_timestamp[tmpthrd] = event->timestamp;
	thread_prev_state[tmpthrd] = prevstate;
	if ((prevstate != THREAD_WAITING_MUTEX) &&
	    (prevstate != THREAD_WAITING_CONDITION) &&
	    (prevstate != THREAD_WAITING_JOIN) &&
	    (prevstate != THREAD_EXIT) &&
	    (prevstate != THREAD_TERMINATE)) {
	    thread_state[tmpthrd] = THREAD_READY;
	}
	break;
      case DM_MONITOR_EXIT:
	lifetimeChangeState(0, event->timestamp, thread_state[0]);
	thread_prev_state[0] = thread_state[0];
	thread_state[0] = THREAD_EXIT;
      default:
	break;
    }
}

