
#ifndef _TUXIAN_H_
#define _TUXIAN_H_


#define MAX_NUM_THREADS_INIT  64
#define MAX_NUM_THREADS_INCR  64
#define MAX_NUM_MUTEXES_INIT  64
#define MAX_NUM_MUTEXES_INCR  64
#define MAX_NUM_CONDITIONS_INIT  32
#define MAX_NUM_CONDITIONS_INCR  32

/**********************************************************************
 *   class:  Tuxian
 **********************************************************************/

class Tuxian {
  private:
    unsigned long	*thread_ids;
    int			max_num_threads;
    int			num_threads;

  protected:
    unsigned long  	init_timestamp;
    unsigned long  	timestamp_resolution;
    double         	timestamp_resolution_inv;

  public:
    Tuxian();
    virtual ~Tuxian() { delete(thread_ids); };
    virtual void create(char*) = 0;
    virtual void update(char*) = 0;
    int threadIndex(unsigned long);
    unsigned long  threadID(int idx) { return thread_ids[idx]; };
    int storeThreadID(unsigned long);
};


/**********************************************************************
 *   subclasses derived from Tuxian
 **********************************************************************/

/*  enum TuxianType {
 *    TX_ThreadLifeTime,
 *    TX_ConditionView,
 *  };
 */


/**********************************************************************
 *   class:  ThreadMap
 **********************************************************************/

class ThreadMap: public Tuxian {
  private:
  protected:
    virtual void threadmapInitialize() = 0;
    virtual void threadmapNewThread(int, unsigned long, unsigned long,
				    int, char*)
      = 0;
    virtual void threadmapNewName(int, char*) = 0;
    virtual void threadmapThreadExit(int, unsigned long, unsigned long)
      = 0;

  public:
    ThreadMap() : Tuxian() {};
    virtual ~ThreadMap() {};
    void create(char*);
    void update(char*);
};


/**********************************************************************
 *   class:  ThreadLifeTime
 **********************************************************************/

enum ThreadStateType {
    THREAD_READY,
    THREAD_RUNNING,
    THREAD_WAITING_MUTEX,
    THREAD_WAITING_CONDITION,
    THREAD_WAITING_JOIN,
    THREAD_END_JOIN,
    THREAD_EXIT,
    THREAD_TERMINATE,
    THREAD_PERTURBATION,
    THREAD_NO_STATE           /* initial, undefined value */
};

class ThreadLifeTime: public Tuxian {
  private:
    int              *parent_thrd_idx;
    unsigned long    *prev_timestamp;
    ThreadStateType  *thread_state, *thread_prev_state;
    char             *thread_detach_flag;
    int              max_num_threads;
    int              ts_tolerance;

  protected:
    virtual void lifetimeInitialize() = 0;
    virtual void lifetimeNewThread(int, unsigned long, int, char*) = 0;
    virtual void lifetimeJoinThread(int, unsigned long, int) = 0;
    virtual void lifetimeChangeState(int, unsigned long, ThreadStateType) = 0;
    
  public:
    ThreadLifeTime(int tlr = 0);
    virtual ~ThreadLifeTime();
    void create(char*);
    void update(char*);
    ThreadStateType getThreadState(int indx) {
	return thread_state[indx];
    }
    unsigned long   getPrevTimestamp(int indx) {
	return prev_timestamp[indx];
    }
};

#endif
