/**********************************************************************
 **                                                                  **
 **                      mission_memory_manager.h                    **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  Copyright 2003 - 2005 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef MISSION_MEMORY_MANAGER_H
#define MISSION_MEMORY_MANAGER_H

/* $Id: mission_memory_manager.h,v 1.1.1.1 2006/07/12 13:37:53 endo Exp $ */

#include <vector>
#include <string>

#include "cbrplanner_protocol.h"
#include "cbrplanner_domain.h"
#include "memory_manager.h"

class MissionMemoryManager : public MemoryManager {

public:
    typedef struct Coordinates_t
    {
        double x;
        double y;
    };

    typedef struct MissionPlan_t
    {
        int type;
        int rating;
        string archType;
        vector<CBRPlanner_PlanSequence_t> plans; // Size of this vector indicates # of robots
        vector<string> maxVelValues; // Size of this vector indicates # of robots also
        vector<string> baseVelValues; // Size of this vector indicates # of robots also
        vector<int> robotIDs; // Size of this vector indicates # of robots also
        vector<int> dataIndexList;
    };

    typedef vector<MissionPlan_t> MissionPlanList_t;

protected:
    MissionPlanList_t missionPlanList_;
    
    static const string CBRLIBRARY_EXTENSION_;
    static const string CBL_STRING_DATA_;
    static const string CBL_STRING_TYPE_;
    static const string CBL_STRING_RATING_;
    static const string CBL_STRING_ARCHTYPE_;
    static const string CBL_STRING_PLAN_;
    static const string CBL_STRING_OPERATOR_;
    static const string CBL_STRING_OPERATOR_NAME_;
    static const string CBL_STRING_OPERATOR_INDEX_;
    static const string CBL_STRING_OPERATOR_DESC_;
    static const string CBL_STRING_OPERATOR_PARM_;
    static const string CBL_STRING_OPERATOR_PARM_NAME_;
    static const string CBL_STRING_OPERATOR_PARM_VALUE_;
    static const string CBL_STRING_OPERATOR_PRECONDITION_;
    static const string CBL_STRING_OPERATOR_PRECONDITION_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_;
    static const string CBL_STRING_OPERATOR_EFFECT_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_INDEX_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_VALUE_;
    static const string CBL_STRING_DECISIONTREE_;
    static const string CBL_STRING_NODE_;
    static const string CBL_STRING_NODE_NAME_;
    static const string CBL_STRING_NODE_LEAF_;
    static const string CBL_STRING_NODE_EDGE_;
    static const string CBL_STRING_NODE_EDGE_PARENT_;
    static const string CBL_STRING_NODE_EDGE_VALUE_;
    static const string CBL_STRING_NODE_EDGE_NEXTNODE_;
    static const string CBL_STRING_NODE_DATA_;
    static const string CBL_STRING_NODE_DATA_INDEXES_;
    static const string CBL_STRING_NODE_DATA_INDEX_;
    static const string CBL_STRING_NODE_DATA_STATUSES_;
    static const string CBL_STRING_NODE_DATA_STATUS_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_NAME_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_VALUE_;
    static const string CBL_STRING_START_;
    static const string CBL_STRING_END_;

    MissionPlan_t loadMissionPlan_(FILE *in);
    DecisionTree_t *loadDecisionTree_(FILE *in);
    DecisionTreeData_t loadNodeData_(FILE *in);
    DecisionTreeNode_t *loadNode_(FILE *in);
    string askCBRLibraryName_(void);
    string addCBRLibraryExtension_(string cbrlibraryName);
    string createCBRLibraryNamePrefix_(void);
    bool saveVerify_(string cbrlibraryName);
    bool loadVerify_(string cbrlibraryName);
    bool hasCBRLibraryExtension_(string cbrlibraryName);
    bool confirmClear_(void);
    void saveCBRLibrary_(string cbrlibraryName);
    void loadCBRLibrary_(FILE *in);
    void checkSaveCBRLibrary(void);
    void writeMissionPlans_(FILE *out);
    void writeDecisionTree_(FILE *out, DecisionTree_t *decisionTree);
    void writeNode_(FILE *out, DecisionTreeNode_t *node);
    void writeNodeData_(FILE *out, DecisionTreeData_t data);

public:
    MissionMemoryManager(void);
    MissionMemoryManager(MissionMemoryManager **self);
    ~MissionMemoryManager(void);

    vector<MissionPlan_t> getSavedMissionPlans(Features_t features, int maxRating);
    int saveNewMissionPlan(MissionPlan_t missionplan, Features_t features);
    bool confirmClear(void);
    void loadCBRLibrary(void);
    void loadCBRLibrary(string cbrlibraryName);
    void saveCBRLibrary(void);
    void confirmSaveCBRLibrary(void);
};

inline bool MissionMemoryManager::confirmClear(void)
{
    return (confirmClear_());
}

#endif
/**********************************************************************
 * $Log: mission_memory_manager.h,v $
 * Revision 1.1.1.1  2006/07/12 13:37:53  endo
 * MissionLab 7.0
 *
 * Revision 1.5  2006/05/15 00:54:55  endo
 * gcc-3.4 upgrade
 *
 * Revision 1.4  2006/02/19 17:52:00  endo
 * Experiment related modifications
 *
 * Revision 1.3  2005/07/31 03:41:41  endo
 * Robot ID specified by CNP can be now recognized by mlab during execution.
 *
 * Revision 1.2  2005/05/18 21:22:22  endo
 * AuRA.naval added.
 *
 * Revision 1.1  2005/02/07 19:53:44  endo
 * Mods from usability-2004
 *
 **********************************************************************/
