/**********************************************************************
 **                                                                  **
 **                             cbrwizard.h                          **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This is a header file for cbrwizard.cc.                         **
 **                                                                  **
 **  Copyright 2006 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef CBRWIZARD_H
#define CBRWIZARD_H

/* $Id: cbrwizard.h,v 1.1.1.1 2006/07/12 13:37:54 endo Exp $ */

#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <Xm/Xm.h>
#include <vector>
#include <string>

#include "symbol.hpp"
#include "load_rc.h"
#include "mission_expert_types.h"
#include "cbrclient.h"

using std::vector;
using std::string;

class CBRWizard {

    typedef struct GeneralYesNoWindowCallbackData_t {
        bool result;
        CBRWizard *cbrWizardInstance;
    };

    typedef struct GeneralToggleWindowCallbackData_t {
        int result;
        CBRWizard *cbrWizardInstance;
        Widget generalToggleWindow;
    };

protected:
    Display *display_;
    XtAppContext appContext_;
    Pixel parentWidgetFg_;
    Pixel parentWidgetBg_;
    Widget parentWidget_;
    Widget generalMessageWindow_;
    symbol_table<rc_chain> rcTable_;
    CBRClient *cbrclient_;
    vector<MExpFeature_t> defaultGlobalFeatures_;
    vector<MExpFeature_t> defaultLocalFeatures_;
    vector<MExpMissionTask_t> defaultTaskList_;
    vector<MExpRobotConstraints_t> robotConstraintsList_;
    string featureFileName_;
    string featureStatusFileName_;
    string overlayFileName_;
    string defaultTaskName_;
    string runtimeCNPMissionManagerTaskName_;
    string runtimeCNPBidderTaskName_;
    double defaultTaskWeight_;
    int cnpMode_;
    int globalFeatureRobotNumIndex_;
    int numTaskTypes_;
    int generalToggleWindowResult_;
    bool generalYesNoWindowResult_;
    bool generalYesNoWindowDone_;
    bool generalToggleWindowDone_;
    bool generalMessageWindowIsUp_;
    bool cbrClientInstantiated_;

    static const struct timespec EVENT_WAITER_SLEEP_TIME_NSEC_;
    static const string EMPTY_STRING_;
    static const string STRING_NUM_ROBOTS_;
    static const string GENERALTOGGLEWINDOW_TITLE_;
    static const double DEFAULT_TASK_WEIGHT_;
    static const int DEFAULT_GENERALTOGGLEWINDOW_WIDTH_;
    static const int DEFAULT_GENERALTOGGLEWINDOW_LINEHEIGHT_;

    Widget createGeneralYesNoWindow_(string question);
    Widget createGeneralYesNoWindow_(const char *question);
    void popupGeneralYesNoWindow_(Widget w);
    void sendGeneralYesNoWindowResult_(bool result);
    void createAndPopupGeneralToggleWindow_(string title, vector<string> options);
    void sendGeneralToggleWindowResult_(int result);
    void createAndPopupGeneralMessageWindow_(string message);
    void popdownGeneralMessageWindow_(void);
    bool confirmUser_(string phrase, bool useOKOnly);
    bool loadFeaturesFromRCFile_(void);

    static void cbGeneralYesNoWindow_(
        Widget w,
        XtPointer client_data,
        XtPointer callback_data);
    static void cbGeneralToggleWindow_(
        Widget w,
        XtPointer client_data,
        XtPointer callback_data);
    static void cbGeneralMessageWindow_(
        Widget w,
        XtPointer client_data,
        XtPointer callback_data);

    virtual bool initCBRClient_(void);

public:
    CBRWizard(void);
    CBRWizard(
        Display *display,
        Widget parent,
        XtAppContext app,
        int initCNPMode);
    virtual ~CBRWizard(void);
};

inline bool CBRWizard::initCBRClient_(void) {return false;}

#endif
/**********************************************************************
 * $Log: cbrwizard.h,v $
 * Revision 1.1.1.1  2006/07/12 13:37:54  endo
 * MissionLab 7.0
 *
 * Revision 1.2  2006/05/14 07:39:22  endo
 * gcc-3.4 upgrade
 *
 * Revision 1.1  2006/01/30 02:53:50  endo
 * AO-FNC CBR-CNP Type-I check-in.
 *
 **********************************************************************/
