/**********************************************************************
 **                                                                  **
 **                         TwoPointsType.hpp                        **
 **                                                                  **
 **                                                                  **
 **  Written by:  Douglas C. MacKenzie                               **
 **                                                                  **
 **  Copyright 1995, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **********************************************************************/

/* $Id: TwoPoints.hpp,v 1.1.1.1 2006/07/12 13:37:59 endo Exp $ */

#ifndef TwoPointsType_HPP
#define TwoPointsType_HPP

#include <unistd.h>

#include <X11/X.h>
#include <Xm/Xm.h>
#include <Xm/Text.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#include <Xm/PanedW.h>
#include <Xm/TextF.h>
#include <Xm/SelectioB.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>

#include "utilities.h"
#include "gt_std.h"
#include "TypeInfo.hpp"

extern bool SHOW_SLIDER_VALUES;

static Widget two_pt_wy;
static Widget two_pt_wx;

void run_ovl_pick_pt_button(void);
int Read_Point_From_Overlay(float *x, float *y);

/*-----------------------------------------------------------------------*/
/* this function is to read the coordinates from overlay by running mlab */

/*-----------------------------------------------------------------------*/
// Define the TwoPointsType class
class TwoPointsType : public TypeInfo
{
   Widget wx;
   Widget wy;

public:

   TwoPointsType(const char *TypeName) : TypeInfo(TypeName)
   {
      wx = NULL;
      wy = NULL;
   }

   ~TwoPointsType() 
   { 
   }

   Widget 
   CreateDisplayWidget(Widget parent, Symbol *parm)
   {
      Widget w, xlabel, ylabel, ovl_pick_pt_button;

      char *label = "";
      if( parm->name )
      {
         if( parm->name[0] == '%' )
            label = &parm->name[1];
         else
            label = parm->name;
      }

      // Get the value
      char *xval;
      if( parm->input_generator && parm->input_generator->name )
      {
	 xval = strdup(parm->input_generator->name);
      }
      else
      {
	 xval = strdup("0,0");
      }

      char *yval = strchr(xval,',');
      if( yval == NULL )
      {
	 free(xval);
	 xval = strdup("0");
	 yval = "0";
      }
      else
      {
	  // Right now, xval is "x.xx, y.yy" and yval is ", y.yy". Replace, ","
	  // of yval with NULL terminate, so that, xval string ends before ",".
	  *yval = '\0';
	  yval++;

	  // Trim any leading whitespace of yval.
	  while((*yval)&&(isspace(*yval))) yval++;
      }

      // Create a blank line
//      XmString str = XmStringCreateSimple(" ");
      XmString str = XmStringCreateLocalized(" ");
      XtVaCreateManagedWidget("",
         	xmLabelGadgetClass, parent,
         	XmNlabelString, str,
         	NULL);
      XmStringFree(str);

/*
      // Create the frame widget
      Widget w = XtVaCreateWidget("", xmFrameWidgetClass, parent,
                NULL);
*/              
      // Create the frame widget
      w = XtVaCreateWidget("", xmFormWidgetClass, parent,
                NULL);
               
      // Create the X text widget
      wx = XtVaCreateManagedWidget("", xmTextWidgetClass, w,
                XmNvalue,   xval,
//		XmNleftAttachment, XmATTACH_WIDGET,
//		XmNleftWidget, xlabel,
		XmNrightAttachment, XmATTACH_FORM,
		XmNtopAttachment, XmATTACH_FORM,
                NULL);

      // Create the Y text widget
      wy = XtVaCreateManagedWidget("", xmTextWidgetClass, w,
                XmNvalue,   yval,
//		XmNleftAttachment, XmATTACH_WIDGET,
//		XmNleftWidget, ylabel,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNtopWidget, wx,
		XmNrightAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
                NULL);

   // Create the ovl_pick_pt_button.
      str = XmStringCreateSimple("Pick from Overlay");
      ovl_pick_pt_button = XtVaCreateManagedWidget(
                "", xmPushButtonGadgetClass, w,
                XmNlabelString, str,
                XmNtopAttachment,     XmATTACH_FORM,
                XmNbottomAttachment,  XmATTACH_FORM,
                XmNleftAttachment,    XmATTACH_FORM,
                XmNshowAsDefault,     False,
                XmNdefaultButtonShadowThickness,1,
                NULL);
      XmStringFree(str);
      XtAddCallback(ovl_pick_pt_button, XmNactivateCallback,
		(XtCallbackProc) run_ovl_pick_pt_button, NULL);
 
      xlabel = XtVaCreateManagedWidget("X",
         	xmLabelGadgetClass, w,
		XmNtopAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_WIDGET,
		XmNrightWidget, wx,
		XmNleftWidget, ovl_pick_pt_button,
                XmNtopOffset,        10,
         	NULL);

      ylabel = XtVaCreateManagedWidget("Y",
         	xmLabelGadgetClass, w,
		XmNtopWidget, xlabel,
		XmNtopAttachment, XmATTACH_WIDGET,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_WIDGET,
		XmNrightWidget, wy,
		XmNleftWidget, ovl_pick_pt_button,
                XmNtopOffset,        10,
         	NULL);

      two_pt_wx = wx;
      two_pt_wy = wy;

      XtManageChild(w);
               
      return w;
   }

   void
   UpdateFromDisplayWidget(Widget w, char **value)
   {
      char *xvalue = NULL;
      char *yvalue = NULL;
      XtVaGetValues(wx,
		    XmNvalue, &xvalue,
		    NULL);

      XtVaGetValues(wy,
		    XmNvalue, &yvalue,
		    NULL);

      // Trim any leading whitespace of xvalue and yvalue.
      while((*xvalue)&&(isspace(*xvalue))) xvalue++;
      while((*yvalue)&&(isspace(*yvalue))) yvalue++;

      char buf[2048];
      sprintf(buf,"%s, %s", xvalue, yvalue);
      *value = strdup(buf);
   }


   char *
   PrintValue(struct Symbol *parm, const char *separator)
   {
      char buf[2048];
      buf[0] = '\0';  

      char *val = "Unknown, Unknown";

      // Get the value
      if( parm->input_generator && parm->input_generator->name )
         val = parm->input_generator->name;

      if( parm->name[0] == '%' )
         sprintf(buf,"%s <%s>", &parm->name[1], val);
      else
         sprintf(buf,"%s <%s>", parm->name, val);
  
      return strdup(buf);
   }
};
/*-----------------------------------------------------------------------*/

void run_ovl_pick_pt_button()
{
  float x, y;
  static int a;

  a = Read_Point_From_Overlay(&x, &y);
  
  if (a==0)
  {
    char xstr[80];
    char ystr[80];

    sprintf( xstr, "%.2f", x);
    sprintf( ystr, "%.2f", y);
    XtUnmanageChild(two_pt_wx);
    XtUnmanageChild(two_pt_wy);

    XmTextSetString( two_pt_wx, xstr);
    XmTextSetString( two_pt_wy, ystr);

    XtManageChild(two_pt_wx);
    XtManageChild(two_pt_wy);
  }
}

int
Read_Point_From_Overlay(float *x, float *y)
{
  FILE* F;
  char tmpfile[40];
  char cmd[80];
  pid_t pid;

  pid = getpid(); //get the current process id to put into the file name
  sprintf(tmpfile, "/tmp/tmp_coor_%d.tmp", pid); //file name
  
  sprintf(cmd, "mlab -m %s -n", tmpfile);
  system(cmd);

  if ( (F = fopen(tmpfile, "r")) == NULL )
  {
    //can't open this file
    unlink(tmpfile); // delete the tmpfile
    return -1;

  } 
  else 
  {
    fscanf(F, "%f, %f", x, y);
    fclose(F);
    unlink(tmpfile); // delete the tmpfile
    return 0;
  }
}

#endif


///////////////////////////////////////////////////////////////////////
// $Log: TwoPoints.hpp,v $
// Revision 1.1.1.1  2006/07/12 13:37:59  endo
// MissionLab 7.0
//
// Revision 1.1.1.1  2005/02/06 23:00:06  endo
// AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
//
// Revision 1.8  2003/04/06 12:58:48  endo
// gcc 3.1.1
//
// Revision 1.7  2002/01/13 01:24:31  endo
// list ->GTList
//
// Revision 1.6  2000/10/16 19:28:28  endo
// Modified due to the compiler upgrade.
//
// Revision 1.5  2000/04/13 22:31:17  endo
// Checked in for Doug.
//
// Revision 1.4  1999/07/03 22:04:03  endo
// Pick-point-from-Overlay feature added.
//
// Revision 1.3  1996/10/04 21:10:59  doug
// changes to get to version 1.0c
//
// Revision 1.3  1996/10/01 13:01:32  doug
// got library writes working
//
// Revision 1.2  1996/05/14  23:17:39  doug
// added twopoints case
//
// Revision 1.1  1996/05/14  18:56:27  doug
// Initial revision
//
///////////////////////////////////////////////////////////////////////
